function C = CProp(P,M,I,X,Y,Z)
%function C = CProp(P,M,I,X,Y,Z)
%
%    Assigns material parameters, C
%
%	P - contains elements nx, ny, and nz (and dz if M = 0)
%	M - Material to assign to C
%	  Options are
%		0 - PREM, Isotropic only
%		1 - Olivine
%		2 - User defined given P.Svel, P.Pvel, and P.Den
%	I - isotropic (0) or anisotropic (1)
%	X, Y, Z - vectors of elements to assign. Specify ':' for all
%	  elements
%
%	Companion script for Oliver S. Boyd, An Efficient Matlab Script to Calculate 
%	Heterogeneous Anisotropically Elastic Wave Propagation in Three Dimensions
%

C = zeros(P.nx,P.ny,P.nz,3,3,3,3);
switch M
    case 0
	if I > 0
		disp('Anisotropic PREM not supported')
		disp('Returning isotropic PREM')
	end
	zvec = P.dz:P.dz:P.dz*P.nz;
	EP = EarthParam(zvec/1e3);
	l = EP.BulkMod(Z) - (2/3)*EP.ShearMod(Z);
	u = EP.ShearMod(Z);
        j = find(u == 0);
        jj = find(u > 0);
        l(j) = min(l(jj));
        u(j) = min(u(jj));
	Lt = ones(P.ny,1)*l;
	Ut = ones(P.ny,1)*u;
	for j = 1:P.nx
		L(j,:,:) = Lt*1e9;
		U(j,:,:) = Ut*1e9;
	end
	C(X,Y,Z,1,1,1,1) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,2,2,2,2) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,3,3,3,3) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,2,2,1,1) = L(X,Y,Z);
	C(X,Y,Z,1,1,2,2) = L(X,Y,Z);
	C(X,Y,Z,3,3,2,2) = L(X,Y,Z);
	C(X,Y,Z,2,2,3,3) = L(X,Y,Z);
	C(X,Y,Z,3,3,1,1) = L(X,Y,Z);
	C(X,Y,Z,1,1,3,3) = L(X,Y,Z);
	C(X,Y,Z,2,3,2,3) = U(X,Y,Z);
	C(X,Y,Z,3,2,2,3) = U(X,Y,Z);
	C(X,Y,Z,2,3,3,2) = U(X,Y,Z);
	C(X,Y,Z,3,2,3,2) = U(X,Y,Z);
	C(X,Y,Z,1,3,1,3) = U(X,Y,Z);
	C(X,Y,Z,3,1,1,3) = U(X,Y,Z);
	C(X,Y,Z,1,3,3,1) = U(X,Y,Z);
	C(X,Y,Z,3,1,3,1) = U(X,Y,Z);
	C(X,Y,Z,1,2,1,2) = U(X,Y,Z);
	C(X,Y,Z,1,2,2,1) = U(X,Y,Z);
	C(X,Y,Z,2,1,1,2) = U(X,Y,Z);
	C(X,Y,Z,2,1,2,1) = U(X,Y,Z);
    case 1
	if I > 0
		%Olivine, orthorhombic
		C(X,Y,Z,1,1,1,1) = 328e9;
		C(X,Y,Z,2,2,2,2) = 200e9;
		C(X,Y,Z,3,3,3,3) = 235e9;
		C(X,Y,Z,2,2,1,1) = 69e9;
		C(X,Y,Z,1,1,2,2) = 69e9;
		C(X,Y,Z,3,3,2,2) = 73e9;
		C(X,Y,Z,2,2,3,3) = 73e9;
		C(X,Y,Z,3,3,1,1) = 69e9;
		C(X,Y,Z,1,1,3,3) = 69e9;
		C(X,Y,Z,2,3,2,3) = 66.7e9;
		C(X,Y,Z,3,2,2,3) = 66.7e9;
		C(X,Y,Z,2,3,3,2) = 66.7e9;
		C(X,Y,Z,3,2,3,2) = 66.7e9;
		C(X,Y,Z,1,3,1,3) = 81.3e9;
		C(X,Y,Z,3,1,1,3) = 81.3e9;
		C(X,Y,Z,1,3,3,1) = 81.3e9;
		C(X,Y,Z,3,1,3,1) = 81.3e9;
		C(X,Y,Z,1,2,1,2) = 80.9e9;
		C(X,Y,Z,1,2,2,1) = 80.9e9;
		C(X,Y,Z,2,1,1,2) = 80.9e9;
		C(X,Y,Z,2,1,2,1) = 80.9e9;
	else
		%Olivine, isotropic
		C(X,Y,Z,1,1,1,1) = 254.3e9;
		C(X,Y,Z,2,2,2,2) = 254.3e9;
		C(X,Y,Z,3,3,3,3) = 254.3e9;
		C(X,Y,Z,2,2,1,1) = 101.7e9;
		C(X,Y,Z,1,1,2,2) = 101.7e9;
		C(X,Y,Z,3,3,2,2) = 101.7e9;
		C(X,Y,Z,2,2,3,3) = 101.7e9;
		C(X,Y,Z,3,3,1,1) = 101.7e9;
		C(X,Y,Z,1,1,3,3) = 101.7e9;
		C(X,Y,Z,2,3,2,3) = 76.3e9;
		C(X,Y,Z,3,2,2,3) = 76.3e9;
		C(X,Y,Z,2,3,3,2) = 76.3e9;
		C(X,Y,Z,3,2,3,2) = 76.3e9;
		C(X,Y,Z,1,3,1,3) = 76.3e9;
		C(X,Y,Z,3,1,1,3) = 76.3e9;
		C(X,Y,Z,1,3,3,1) = 76.3e9;
		C(X,Y,Z,3,1,3,1) = 76.3e9;
		C(X,Y,Z,1,2,1,2) = 76.3e9;
		C(X,Y,Z,1,2,2,1) = 76.3e9;
		C(X,Y,Z,2,1,1,2) = 76.3e9;
		C(X,Y,Z,2,1,2,1) = 76.3e9;
	end
    case 2
	u = P.Den.*P.Svel.^2;
	l = P.Den.*P.Pvel.^2 - 2*u;
	Lt = ones(P.ny,1)*l;
	Ut = ones(P.ny,1)*u;
	for j = 1:P.nx
		L(j,:,:) = Lt;
		U(j,:,:) = Ut;
	end
	C(X,Y,Z,1,1,1,1) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,2,2,2,2) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,3,3,3,3) = L(X,Y,Z) + 2*U(X,Y,Z);
	C(X,Y,Z,2,2,1,1) = L(X,Y,Z);
	C(X,Y,Z,1,1,2,2) = L(X,Y,Z);
	C(X,Y,Z,3,3,2,2) = L(X,Y,Z);
	C(X,Y,Z,2,2,3,3) = L(X,Y,Z);
	C(X,Y,Z,3,3,1,1) = L(X,Y,Z);
	C(X,Y,Z,1,1,3,3) = L(X,Y,Z);
	C(X,Y,Z,2,3,2,3) = U(X,Y,Z);
	C(X,Y,Z,3,2,2,3) = U(X,Y,Z);
	C(X,Y,Z,2,3,3,2) = U(X,Y,Z);
	C(X,Y,Z,3,2,3,2) = U(X,Y,Z);
	C(X,Y,Z,1,3,1,3) = U(X,Y,Z);
	C(X,Y,Z,3,1,1,3) = U(X,Y,Z);
	C(X,Y,Z,1,3,3,1) = U(X,Y,Z);
	C(X,Y,Z,3,1,3,1) = U(X,Y,Z);
	C(X,Y,Z,1,2,1,2) = U(X,Y,Z);
	C(X,Y,Z,1,2,2,1) = U(X,Y,Z);
	C(X,Y,Z,2,1,1,2) = U(X,Y,Z);
	C(X,Y,Z,2,1,2,1) = U(X,Y,Z);
    case 1
end
