function [u, f] = FDWave(P)
%function [u, f] = FDWave(P)
%
%    Calculates 3d, 2d and 1d finite difference solution to wave equation - 
%		all variables must have same units, e.g. mks, cgs, etc.
%	INPUT
%       P - structure containing solution parameters.  This structure defines the following variables
%		nx,dx - number of x elements and distance between x elements
%		ny,dy - number of y elements and distance between y elements
%		nz,dz - number of z elements and distance between z elements
%		nt,dt - number of t elements and distance between t elements
%		C(3,3,3,3,nx,ny,nz) Full aniostropic tensor
%		D(nx,ny,nz) - density
%		T - Dominant period or source wavelet
%		S - incoming slowness of source wavelet
%		N - damping factor related to Q, the quality factor, by N=C/wQ where w is the angular frequency
%
%	   Boundary conditions on the six faces of the model domain.  These variables may have the values of:
%	     0 - Fixed boundary
%	     1 - Free Surface
%	     2 - Absorbing according to higdon, 1991 with m = 1 and user specified velocity, AbVel
%		BCxy1 (Applied in 1D, 2D, and 3D)
%		BCxy2 (Applied in 1D, 2D, and 3D)
%		BCyz1 (Applied in 2D, and 3D)
%		BCyz2 (Applied in 2D, and 3D)
%		BCxz1 (Applied in 3D)
%		BCxz2 (Applied in 3D)
%
%		AbVel - Velocity used for absorbing boundary condition on the horizontal faces.
%		AbVel2 - Velocity used for absorbing boundary condition on the vertical faces.
%		ITAx,ITAy,ITAz - Intial Time series displacements applied at the following element positions. 
%		   It must have the same dt as the model and be of size (ITx,ITy,ITz,ITnt).
%		ITx - IT x positions
%		ITy - IT y positions
%		ITz - IT z positions
%		ITnt - IT number of samples
%		   For example, IT might be a sine wave.  ITx could be 1:nx, ITy could be 1:ny, and ITz could be
%		   nz.  This would apply IT at the xy face located at nz.
%		ud - vector of depths to examine ux, uy, and uz
%		   If length(ud) > 1, this program will return all amplitudes at ud for all x's at ny/2 for all t's.
%		   Otherwise, this program will return all amplitudes at ud for all x's and y's for all t's
%		nf - number of frames to capture; if equal to 0, no figure will be displayed and no frames captured.
%		fig - figure to use to display wavefield.
%		X - amplification factor for the x-displacement in the figure
%		Y - amplification factor for the y-displacement in the figure
%		Z - amplification factor for the z-displacement in the figure
%	u - structure containing displacements ux, uy, and uz as functions of
%	    time for vector of depths given by ud.
%	f - frames that can be used by movie(f) to show the evolution of the
%	    wavefield. Frames are acquired for xz profiles every 10 times steps.
%
%	NOTE: To maintain numerical stability, 
%	      P.dt <= P.dx|P.dy|P.dz/sqrt(P.U/P.D + (P.L+2*P.U)/P.D);
%
%             To reduce grid dispersion, The wavelength at the upper half-power frequency
%	      should be 10 times the number of grid points. For example, if the wavelength
%             is 10 km, the grid spacing should be less than 1 km (Alford et al. 1974).
%
%	Reference:
%	Oliver S. Boyd
%	An Efficient Matlab Script to Calculate Heterogeneous Anisotropically 
%	Elastic Wave Propagation in Three Dimensions, Computers and Geosciences, 32, 259-264, 2006
%
%	Updated 6/1/08 to include attenuation and reduce execution time.
%	Updated 8/25/08 to include reduce execution time for 1d and 2d cases.
%	Updated 12/10/08 to include periodic boundary conditions
%


%
%	Explanation of FDWave.m
%
%	Each subroutine starts with
%
%	%%%%%%
%	%%%%%%
%	%%%%%%
%
%	There are five subroutines.
%
%	FD_3D
%	Clear
%	Initialize
%	GetDerivatives
%	ApplyBC
%
%	FD_3D - Calls all other subroutines, generates displacement within model and returns output
%		to the user.
%
%	Clear - Clears variables used within program.
%
%	Initialize - Initializes matrices used within program to reduce time for allocation.
%
%	GetMatDerivatives - Calculates material derivatives within model space.
%
%	GetDerivatives - Calculates derivatives within model space for each time step.
%
%	ApplyBC - Applies boundary conditions.
%
      %d1txx d1txy d1txz ...
      %d1tyx d1tyy d1tyz ...
      %d1tzx d1tzy d1tzz ...

global u f fr ...
       ux1 ux2 ux3 ...
       uy1 uy2 uy3 ...
       uz1 uz2 uz3;

Clear

if P.nf > 0
   figure(P.fig); clf
   set(gca,'position',[0.1 0.1 0.8 0.7])
   skip = round(P.nt/P.nf);
else
   skip = round(P.nt/30);
   % If you don't want any output, set skip to greater than P.nt
end

%
%	Insure proper starting conditions in terms of numerical stability and dispersion.
%

if P.nx == 1
	sVs = [P.C(:,2,3,2,3) +  P.C(:,1,3,1,3) +  P.C(:,1,2,1,2)]/3;
	sVp = [P.C(:,1,1,1,1) +  P.C(:,2,2,2,2) +  P.C(:,3,3,3,3)]/3;
elseif P.ny == 1
	sVs = [P.C(:,:,2,3,2,3) +  P.C(:,:,1,3,1,3) +  P.C(:,:,1,2,1,2)]/3;
	sVp = [P.C(:,:,1,1,1,1) +  P.C(:,:,2,2,2,2) +  P.C(:,:,3,3,3,3)]/3;
else
	sVs = [P.C(:,:,:,2,3,2,3) +  P.C(:,:,:,1,3,1,3) +  P.C(:,:,:,1,2,1,2)]/3;
	sVp = [P.C(:,:,:,1,1,1,1) +  P.C(:,:,:,2,2,2,2) +  P.C(:,:,:,3,3,3,3)]/3;
end
sD = squeeze(P.D);
if P.nx == 1
	sD = sD';
end
MaxVs = max(max(max(sqrt(sVs./sD))));
MaxVp = max(max(max(sqrt(sVp./sD))));
MinVel = min(min(min(sqrt(sVs./sD))));
WL = MinVel*P.T;
disp('Dispersion')
disp(['  Dispersion is minimized if there are more than 10 grid point per wavelength'])
disp(['  There are ' num2str(WL/P.dz) ' grid points per wavelength.'])

disp(' ')
disp('Numerical Stability')
Criteria = P.dz./(MaxVs + MaxVp);
%Criteria = max(max(max(P.dx./sqrt(P.U./P.D + (P.L+2*P.U)./P.D))))
if P.dt > Criteria
        disp('  Numerical instability will result.')
        disp(['  dt should be less than ' num2str(Criteria)])
        disp(['  dt is currently ' num2str(P.dt)])
	disp('END')
        return
else
        disp('  Numerical stability will result.')
        disp(['  dt should be less than ' num2str(Criteria)])
        disp(['  dt is currently ' num2str(P.dt)])
end
%       return
disp(' ')

%
%	Begin finite difference calculation
%

%	Initialize
dt2_PD = squeeze((P.dt^2)./P.D);
nd = length(P.ud);
fr = 0;

if P.nx == 1
   Initialize1D(P)
   GetMatDerivatives1D(P)
elseif P.ny == 1
   Initialize2D(P)
   GetMatDerivatives2D(P)
else
   Initialize3D(P)
   GetMatDerivatives3D(P)
end

[dx,dy,dz,P.dITAx] = gradient(P.ITAx,1);
[dx,dy,dz,P.dITAy] = gradient(P.ITAy,1);
[dx,dy,dz,P.dITAz] = gradient(P.ITAz,1);

%	Calculate wave fields
for t = 1:P.nt
   if P.nx == 1 	% 1D case
      GetDerivatives1D(P)
      if P.Iso == 1
         [rhsX, rhsY, rhsZ] = rhs(P,1);
      else
         [rhsX, rhsY, rhsZ] = rhs(P,4);
      end
      if P.Ela == 0
         [rhsXN, rhsYN, rhsZN] = rhs(P,7);
         rhsX = rhsX + rhsXN;
         rhsY = rhsY + rhsYN;
         rhsZ = rhsZ + rhsZN;
      end
   elseif P.ny == 1 	% 2D case
      GetDerivatives2D(P)
      if P.Iso == 1
         [rhsX, rhsY, rhsZ] = rhs(P,2);
      else
         [rhsX, rhsY, rhsZ] = rhs(P,5);
      end
      if P.Ela == 0
         [rhsXN, rhsYN, rhsZN] = rhs(P,8);
         rhsX = rhsX + rhsXN;
         rhsY = rhsY + rhsYN;
         rhsZ = rhsZ + rhsZN;
      end
   else			% 3D case
      GetDerivatives3D(P)
      if P.Iso == 1
         [rhsX, rhsY, rhsZ] = rhs(P,3);
      else
         [rhsX, rhsY, rhsZ] = rhs(P,6);
      end
      if P.Ela == 0
         [rhsXN, rhsYN, rhsZN] = rhs(P,9);
         rhsX = rhsX + rhsXN;
         rhsY = rhsY + rhsYN;
         rhsZ = rhsZ + rhsZN;
      end
   end
   ux3 = dt2_PD.*rhsX + 2*ux2 - ux1;
   uy3 = dt2_PD.*rhsY + 2*uy2 - uy1;
   uz3 = dt2_PD.*rhsZ + 2*uz2 - uz1;

   % Apply boundary conditions
   if P.nx == 1 
      ApplyBC1D(t,P)
   elseif P.ny == 1
      ApplyBC2D(t,P)
   else
      ApplyBC3D(t,P)
   end

   ux1 = ux2;
   ux2 = ux3;
   uy1 = uy2;
   uy2 = uy3;   
   uz1 = uz2;
   uz2 = uz3;
   
   % Set up variables to return to user.
   GetRetVar(nd, t, skip, P);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		CLEAR						%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Clear

global ux1 ux2 ux3 ...
       uy1 uy2 uy3 ...
       uz1 uz2 uz3 ...
       BCyz BCxz BCxy ...
       d1uxx1 d1uxy1 d1uxz1 ...
       d1uyx1 d1uyy1 d1uyz1 ...
       d1uzx1 d1uzy1 d1uzz1 ...
       d2uxxx1 d2uxyx1 d2uxzx1 ...
       d2uyxx1 d2uyyx1 d2uyzx1 ...
       d2uzxx1 d2uzyx1 d2uzzx1 ...
       d2uxxy1 d2uxyy1 d2uxzy1 ...
       d2uyxy1 d2uyyy1 d2uyzy1 ...
       d2uzxy1 d2uzyy1 d2uzzy1 ...
       d2uxxz1 d2uxyz1 d2uxzz1 ...
       d2uyxz1 d2uyyz1 d2uyzz1 ...
       d2uzxz1 d2uzyz1 d2uzzz1 ...
       d1uxx2 d1uxy2 d1uxz2 ...
       d1uyx2 d1uyy2 d1uyz2 ...
       d1uzx2 d1uzy2 d1uzz2 ...
       d2uxxx2 d2uxyx2 d2uxzx2 ...
       d2uyxx2 d2uyyx2 d2uyzx2 ...
       d2uzxx2 d2uzyx2 d2uzzx2 ...
       d2uxxy2 d2uxyy2 d2uxzy2 ...
       d2uyxy2 d2uyyy2 d2uyzy2 ...
       d2uzxy2 d2uzyy2 d2uzzy2 ...
       d2uxxz2 d2uxyz2 d2uxzz2 ...
       d2uyxz2 d2uyyz2 d2uyzz2 ...
       d2uzxz2 d2uzyz2 d2uzzz2 ...
       d2uxxt d2uxyt d2uxzt ...
       d2uyxt d2uyyt d2uyzt ...
       d2uzxt d2uzyt d2uzzt ...
       d3uxxxt d3uxyxt d3uxzxt ...
       d3uyxxt d3uyyxt d3uyzxt ...
       d3uzxxt d3uzyxt d3uzzxt ...
       d3uxxyt d3uxyyt d3uxzyt ...
       d3uyxyt d3uyyyt d3uyzyt ...
       d3uzxyt d3uzyyt d3uzzyt ...
       d3uxxzt d3uxyzt d3uxzzt ...
       d3uyxzt d3uyyzt d3uyzzt ...
       d3uzxzt d3uzyzt d3uzzzt ...
       d1Lx d1Ly d1Lz ...
       d1Ux d1Uy d1Uz ...
       d1Cx d1Cy d1Cz ...
       d1Nx d1Ny d1Nz;

clear ux1 ux2 ux3 ...
      uy1 uy2 uy3 ...
      uz1 uz2 uz3 ...
      BCyz BCxz BCxy ...
      d1uxx1 d1uxy1 d1uxz1 ...
      d1uyx1 d1uyy1 d1uyz1 ...
      d1uzx1 d1uzy1 d1uzz1 ...
      d2uxxx1 d2uxyx1 d2uxzx1 ...
      d2uyxx1 d2uyyx1 d2uyzx1 ...
      d2uzxx1 d2uzyx1 d2uzzx1 ...
      d2uxxy1 d2uxyy1 d2uxzy1 ...
      d2uyxy1 d2uyyy1 d2uyzy1 ...
      d2uzxy1 d2uzyy1 d2uzzy1 ...
      d2uxxz1 d2uxyz1 d2uxzz1 ...
      d2uyxz1 d2uyyz1 d2uyzz1 ...
      d2uzxz1 d2uzyz1 d2uzzz1 ...
      d1uxx2 d1uxy2 d1uxz2 ...
      d1uyx2 d1uyy2 d1uyz2 ...
      d1uzx2 d1uzy2 d1uzz2 ...
      d2uxxx2 d2uxyx2 d2uxzx2 ...
      d2uyxx2 d2uyyx2 d2uyzx2 ...
      d2uzxx2 d2uzyx2 d2uzzx2 ...
      d2uxxy2 d2uxyy2 d2uxzy2 ...
      d2uyxy2 d2uyyy2 d2uyzy2 ...
      d2uzxy2 d2uzyy2 d2uzzy2 ...
      d2uxxz2 d2uxyz2 d2uxzz2 ...
      d2uyxz2 d2uyyz2 d2uyzz2 ...
      d2uzxz2 d2uzyz2 d2uzzz2 ...
      d2uxxt d2uxyt d2uxzt ...
      d2uyxt d2uyyt d2uyzt ...
      d2uzxt d2uzyt d2uzzt ...
      d3uxxxt d3uxyxt d3uxzxt ...
      d3uyxxt d3uyyxt d3uyzxt ...
      d3uzxxt d3uzyxt d3uzzxt ...
      d3uxxyt d3uxyyt d3uxzyt ...
      d3uyxyt d3uyyyt d3uyzyt ...
      d3uzxyt d3uzyyt d3uzzyt ...
      d3uxxzt d3uxyzt d3uxzzt ...
      d3uyxzt d3uyyzt d3uyzzt ...
      d3uzxzt d3uzyzt d3uzzzt ...
      d1Lx d1Ly d1Lz ...
      d1Ux d1Uy d1Uz ...
      d1Cx d1Cy d1Cz ...
      d1Nx d1Ny d1Nz;
      
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		INITIALIZE 1D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Initialize1D(P)

global u f ...
       BCxy ...
       qx qt qxt ...
       prevk curk nextk ...
       ux1 ux2 ux3 ...
       uy1 uy2 uy3 ...
       uz1 uz2 uz3 ...
       d1uxz1 ...
       d1uyz1 ...
       d1uzz1 ...
       d2uxzz1 ...
       d2uyzz1 ...
       d2uzzz1 ...
       d1uxz2 ...
       d1uyz2 ...
       d1uzz2 ...
       d2uxzz2 ...
       d2uyzz2 ...
       d2uzzz2 ...
       d2uxzt ...
       d2uyzt ...
       d2uzzt ...
       d3uxzzt ...
       d3uyzzt ...
       d3uzzzt ...
       d1Lz ...
       d1Uz ...
       d1Cz ...
       d1Nz;
      
%Derivatives of material properties

if P.Iso == 1
	d1Lz = zeros(1,P.nz);
	d1Uz = zeros(1,P.nz);
else
	d1Cz = zeros(P.nz,3,3,3,3);
end
if P.Ela == 0
	d1Nz = zeros(P.nz,3,3,3,3);
end

%Displacements for three time steps

ux1 = zeros(1,P.nz); %Displacement in the x direction for t - 1
ux2 = zeros(1,P.nz); %Displacement in the x direction for t 
ux3 = zeros(1,P.nz); %Displacement in the x direction for t + 1

uy1 = zeros(1,P.nz); %Displacement in the y direction for t - 1
uy2 = zeros(1,P.nz); %Displacement in the y direction for t 
uy3 = zeros(1,P.nz); %Displacement in the y direction for t + 1

uz1 = zeros(1,P.nz); %Displacement in the z direction for t - 1
uz2 = zeros(1,P.nz); %Displacement in the z direction for t 
uz3 = zeros(1,P.nz); %Displacement in the z direction for t + 1

%Derivatives of displacements

d1uxz1 = zeros(1,P.nz); 
d1uyz1 = zeros(1,P.nz); 
d1uzz1 = zeros(1,P.nz); 

d2uxzz1 = zeros(1,P.nz); 
d2uyzz1 = zeros(1,P.nz); 
d2uzzz1 = zeros(1,P.nz); 

d1uxz2 = zeros(1,P.nz); 
d1uyz2 = zeros(1,P.nz); 
d1uzz2 = zeros(1,P.nz); 

d2uxzz2 = zeros(1,P.nz); 
d2uyzz2 = zeros(1,P.nz); 
d2uzzz2 = zeros(1,P.nz); 

d2uxzt = zeros(1,P.nz); 
d2uyzt = zeros(1,P.nz); 
d2uzzt = zeros(1,P.nz); 

d3uxzzt = zeros(1,P.nz); 
d3uyzzt = zeros(1,P.nz); 
d3uzzzt = zeros(1,P.nz); 

%Free surface boundary condition, du = 0
%Fixed surface boundary condition, u = 0

%Absorbing boundary condition from Higdon, 1991, modifications on
%   page 37 of lab notebook. Following are absorbing parameters.

BCxy = zeros(P.nx,P.ny,1);

B = 1;
v = P.AbVel*P.dt/P.dz;
b = .4;
qx = (b*(B + v) - v)/((B + v)*(1 - b));
qt = (b*(B + v) - B)/((B + v)*(1 - b));
qxt = b/(b-1);

%Initialize other variables

prevk = 1:P.nz-2;
curk = 2:P.nz-1;
nextk = 3:P.nz;

%Output parameters

nd = length(P.ud);
u.ux = zeros(nd,P.nt);
u.uy = zeros(nd,P.nt);
u.uz = zeros(nd,P.nt);
if P.nf > 0
   f = zeros(3*P.nx,P.nz,P.nf);
else
   f = [];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETMATDERIVATIVES 1D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetMatDerivatives1D(P)

global prevk curk nextk ...
       d1Lz ...
       d1Uz ...
       d1Cz ...
       d1Nz;

if P.Iso == 1
	d1Lz(curk) = squeeze((P.L(nextk) - P.L(prevk))/(2*P.dz));
	d1Uz(curk) = squeeze((P.U(nextk) - P.U(prevk))/(2*P.dz));
else
	d1Cz(curk,:,:,:,:) = squeeze((P.C(nextk,:,:,:,:) - P.C(prevk,:,:,:,:))/(2*P.dz));
end
if P.Ela == 0
	d1Nz(curk,:,:,:,:) = squeeze((P.N(nextk,:,:,:,:) - P.N(prevk,:,:,:,:))/(2*P.dz));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETDERIVATIVES 1D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetDerivatives1D(P)

global ux1 uy1 uz1 ...
       ux2 uy2 uz2 ...
       ux3 uy3 uz3 ...
       prevk curk nextk ...
       d1uxz1 ...
       d1uyz1 ...
       d1uzz1 ...
       d2uxzz1 ...
       d2uyzz1 ...
       d2uzzz1 ...
       d1uxz2 ...
       d1uyz2 ...
       d1uzz2 ...
       d2uxzz2 ...
       d2uyzz2 ...
       d2uzzz2 ...
       d2uxzt ...
       d2uyzt ...
       d2uzzt ...
       d3uxzzt ...
       d3uyzzt ...
       d3uzzzt;

d1uxz1(curk) = (ux1(nextk) - ux1(prevk))/(2*P.dz);
d1uyz1(curk) = (uy1(nextk) - uy1(prevk))/(2*P.dz);
d1uzz1(curk) = (uz1(nextk) - uz1(prevk))/(2*P.dz);

d2uxzz1(curk) = (ux1(nextk) - 2*ux1(curk) + ux1(prevk))/(P.dz^2);
d2uyzz1(curk) = (uy1(nextk) - 2*uy1(curk) + uy1(prevk))/(P.dz^2);
d2uzzz1(curk) = (uz1(nextk) - 2*uz1(curk) + uz1(prevk))/(P.dz^2);

d1uxz2(curk) = (ux2(nextk) - ux2(prevk))/(2*P.dz);
d1uyz2(curk) = (uy2(nextk) - uy2(prevk))/(2*P.dz);
d1uzz2(curk) = (uz2(nextk) - uz2(prevk))/(2*P.dz);

d2uxzz2(curk) = (ux2(nextk) - 2*ux2(curk) + ux2(prevk))/(P.dz^2);
d2uyzz2(curk) = (uy2(nextk) - 2*uy2(curk) + uy2(prevk))/(P.dz^2);
d2uzzz2(curk) = (uz2(nextk) - 2*uz2(curk) + uz2(prevk))/(P.dz^2);

d2uxzt = (d1uxz2 - d1uxz1)/P.dt;
d2uyzt = (d1uyz2 - d1uyz1)/P.dt;
d2uzzt = (d1uzz2 - d1uzz1)/P.dt;

d3uxzzt = (d2uxzz2 - d2uxzz1)/P.dt;
d3uyzzt = (d2uyzz2 - d2uyzz1)/P.dt;
d3uzzzt = (d2uzzz2 - d2uzzz1)/P.dt;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		APPLYBC 1D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ApplyBC1D(t,P)

global ux2 ux3 ...
       uy2 uy3 ...
       uz2 uz3 ...
       BCxy ...
       qx qt qxt;
      
switch P.BCxy1
	case 0
		ux3(1) = BCxy;
		uy3(1) = BCxy;
		uz3(1) = BCxy;
	case 1
		ux3(1) = ux3(:,2);
		uy3(1) = uy3(:,2);
		uz3(1) = uz3(:,2);
	case 2
		ux3(1) = -qx*ux3(2) + -qt*ux2(1) + -qxt*ux2(2);
		uy3(1) = -qx*uy3(2) + -qt*uy2(1) + -qxt*uy2(2);
		uz3(1) = -qx*uz3(2) + -qt*uz2(1) + -qxt*uz2(2);
end
switch P.BCxy2
	case 0
		ux3(P.nz) = BCxy;
		uy3(P.nz) = BCxy;
		uz3(P.nz) = BCxy;
	case 1
		ux3(P.nz) = ux3(P.nz-1);
		uy3(P.nz) = uy3(P.nz-1);
		uz3(P.nz) = uz3(P.nz-1);
	case 2
		ux3(P.nz) = -qx*ux3(P.nz-1) + -qt*ux2(P.nz) + -qxt*ux2(P.nz-1);
		uy3(P.nz) = -qx*uy3(P.nz-1) + -qt*uy2(P.nz) + -qxt*uy2(P.nz-1);
		uz3(P.nz) = -qx*uz3(P.nz-1) + -qt*uz2(P.nz) + -qxt*uz2(P.nz-1);
end

if t < P.ITnt + 1
	%ux3(P.ITz) = ux3(P.ITz) + P.dITAx(:,t);
	%uy3(P.ITz) = uy3(P.ITz) + P.dITAy(:,t);
	%uz3(P.ITz) = uz3(P.ITz) + P.dITAz(:,t);
	ux3(P.ITz) = P.ITAx(:,t);
	uy3(P.ITz) = P.ITAy(:,t);
	uz3(P.ITz) = P.ITAz(:,t);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		INITIALIZE 2D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Initialize2D(P)

global u f ...
       BCxy BCyz ...
       qx qt qxt ...
       qx2 qt2 qxt2 ...
       previ curi nexti ...
       prevk curk nextk ...
       ux1 ux2 ux3 ...
       uy1 uy2 uy3 ...
       uz1 uz2 uz3 ...
       d1uxx1 d1uxz1 ...
       d1uyx1 d1uyz1 ...
       d1uzx1 d1uzz1 ...
       d2uxxx1 d2uxzx1 ...
       d2uyxx1 d2uyzx1 ...
       d2uzxx1 d2uzzx1 ...
       d2uxxz1 d2uxzz1 ...
       d2uyxz1 d2uyzz1 ...
       d2uzxz1 d2uzzz1 ...
       d1uxx2 d1uxz2 ...
       d1uyx2 d1uyz2 ...
       d1uzx2 d1uzz2 ...
       d2uxxx2 d2uxzx2 ...
       d2uyxx2 d2uyzx2 ...
       d2uzxx2 d2uzzx2 ...
       d2uxxz2 d2uxzz2 ...
       d2uyxz2 d2uyzz2 ...
       d2uzxz2 d2uzzz2 ...
       d2uxxt d2uxzt ...
       d2uyxt d2uyzt ...
       d2uzxt d2uzzt ...
       d3uxxxt d3uxzxt ...
       d3uyxxt d3uyzxt ...
       d3uzxxt d3uzzxt ...
       d3uxxzt d3uxzzt ...
       d3uyxzt d3uyzzt ...
       d3uzxzt d3uzzzt ...
       d1Lx d1Lz ...
       d1Ux d1Uz ...
       d1Cx d1Cz ...
       d1Nx d1Nz;
      
%Derivatives of material properties

if P.Iso == 1
	d1Lx = zeros(P.nx,P.nz);
	d1Lz = zeros(P.nx,P.nz);
	d1Ux = zeros(P.nx,P.nz);
	d1Uz = zeros(P.nx,P.nz);
else
	d1Cx = zeros(P.nx,P.nz,3,3,3,3);
	d1Cz = zeros(P.nx,P.nz,3,3,3,3);
end
if P.Ela == 0
	d1Nx = zeros(P.nx,P.nz,3,3,3,3);
	d1Nz = zeros(P.nx,P.nz,3,3,3,3);
end

%Displacements for three time steps

ux1 = zeros(P.nx,P.nz); %Displacement in the x direction for t - 1
ux2 = zeros(P.nx,P.nz); %Displacement in the x direction for t 
ux3 = zeros(P.nx,P.nz); %Displacement in the x direction for t + 1

uy1 = zeros(P.nx,P.nz); %Displacement in the y direction for t - 1
uy2 = zeros(P.nx,P.nz); %Displacement in the y direction for t 
uy3 = zeros(P.nx,P.nz); %Displacement in the y direction for t + 1

uz1 = zeros(P.nx,P.nz); %Displacement in the z direction for t - 1
uz2 = zeros(P.nx,P.nz); %Displacement in the z direction for t 
uz3 = zeros(P.nx,P.nz); %Displacement in the z direction for t + 1

%Derivatives of displacements

d1uxx1 = zeros(P.nx,P.nz); 
d1uxz1 = zeros(P.nx,P.nz); 
d1uyx1 = zeros(P.nx,P.nz); 
d1uyz1 = zeros(P.nx,P.nz); 
d1uzx1 = zeros(P.nx,P.nz); 
d1uzz1 = zeros(P.nx,P.nz); 

d2uxxx1 = zeros(P.nx,P.nz); 
d2uxzx1 = zeros(P.nx,P.nz); 
d2uyxx1 = zeros(P.nx,P.nz); 
d2uyzx1 = zeros(P.nx,P.nz); 
d2uzxx1 = zeros(P.nx,P.nz); 
d2uzzx1 = zeros(P.nx,P.nz); 
d2uxxz1 = zeros(P.nx,P.nz); 
d2uxzz1 = zeros(P.nx,P.nz); 
d2uyxz1 = zeros(P.nx,P.nz); 
d2uyzz1 = zeros(P.nx,P.nz); 
d2uzxz1 = zeros(P.nx,P.nz); 
d2uzzz1 = zeros(P.nx,P.nz); 

d1uxx2 = zeros(P.nx,P.nz); 
d1uxz2 = zeros(P.nx,P.nz); 
d1uyx2 = zeros(P.nx,P.nz); 
d1uyz2 = zeros(P.nx,P.nz); 
d1uzx2 = zeros(P.nx,P.nz); 
d1uzz2 = zeros(P.nx,P.nz); 

d2uxxx2 = zeros(P.nx,P.nz); 
d2uxzx2 = zeros(P.nx,P.nz); 
d2uyxx2 = zeros(P.nx,P.nz); 
d2uyzx2 = zeros(P.nx,P.nz); 
d2uzxx2 = zeros(P.nx,P.nz); 
d2uzzx2 = zeros(P.nx,P.nz); 
d2uxxz2 = zeros(P.nx,P.nz); 
d2uxzz2 = zeros(P.nx,P.nz); 
d2uyxz2 = zeros(P.nx,P.nz); 
d2uyzz2 = zeros(P.nx,P.nz); 
d2uzxz2 = zeros(P.nx,P.nz); 
d2uzzz2 = zeros(P.nx,P.nz); 

d2uxxt = zeros(P.nx,P.nz); 
d2uxzt = zeros(P.nx,P.nz); 
d2uyxt = zeros(P.nx,P.nz); 
d2uyzt = zeros(P.nx,P.nz); 
d2uzxt = zeros(P.nx,P.nz); 
d2uzzt = zeros(P.nx,P.nz); 

d3uxxxt = zeros(P.nx,P.nz); 
d3uxzxt = zeros(P.nx,P.nz); 
d3uyxxt = zeros(P.nx,P.nz); 
d3uyzxt = zeros(P.nx,P.nz); 
d3uzxxt = zeros(P.nx,P.nz); 
d3uzzxt = zeros(P.nx,P.nz); 
d3uxxzt = zeros(P.nx,P.nz); 
d3uxzzt = zeros(P.nx,P.nz); 
d3uyxzt = zeros(P.nx,P.nz); 
d3uyzzt = zeros(P.nx,P.nz); 
d3uzxzt = zeros(P.nx,P.nz); 
d3uzzzt = zeros(P.nx,P.nz); 

%Free surface boundary condition, du = 0
%Fixed surface boundary condition, u = 0

%Absorbing boundary condition from Higdon, 1991, modifications on
%   page 37 of lab notebook. Following are absorbing parameters.

BCyz = zeros(1,P.ny,P.nz);
BCxy = zeros(P.nx,P.ny,1);

B = 1;
v = P.AbVel*P.dt/P.dz;
b = .4;
qx = (b*(B + v) - v)/((B + v)*(1 - b));
qt = (b*(B + v) - B)/((B + v)*(1 - b));
qxt = b/(b-1);

B = 1;
v = P.AbVel2*P.dt/P.dx;
b = .4;
qx2 = (b*(B + v) - v)/((B + v)*(1 - b));
qt2 = (b*(B + v) - B)/((B + v)*(1 - b));
qxt2 = b/(b-1);

%Initialize other variables

previ = 1:P.nx-2;
curi = 2:P.nx-1;
nexti = 3:P.nx;
prevk = 1:P.nz-2;
curk = 2:P.nz-1;
nextk = 3:P.nz;

%Output parameters

nd = length(P.ud);
if nd > 1
	u.ux = zeros(nd,P.nx,P.nt);
	u.uy = zeros(nd,P.nx,P.nt);
	u.uz = zeros(nd,P.nx,P.nt);
else
	u.ux = zeros(P.nx,P.nt);
	u.uy = zeros(P.nx,P.nt);
	u.uz = zeros(P.nx,P.nt);
end
if P.nf > 0
   f = zeros(3*P.nx,P.nz,P.nf);
else
   f = [];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETMATDERIVATIVES 2D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetMatDerivatives2D(P)

global previ curi nexti ...
       prevk curk nextk ...
       d1Lx d1Lz ...
       d1Ux d1Uz ...
       d1Cx d1Cz ...
       d1Nx d1Nz;

if P.Iso == 1
	d1Lx(curi,:) = squeeze((P.L(nexti,:) - P.L(previ,:))/(2*P.dx));
	d1Lz(:,curk) = squeeze((P.L(:,nextk) - P.L(:,prevk))/(2*P.dz));
	d1Ux(curi,:) = squeeze((P.U(nexti,:) - P.U(previ,:))/(2*P.dx));
	d1Uz(:,curk) = squeeze((P.U(:,nextk) - P.U(:,prevk))/(2*P.dz));
else
	d1Cx(curi,:,:,:,:,:) = squeeze((P.C(nexti,:,:,:,:,:) - P.C(previ,:,:,:,:,:))/(2*P.dx));
	d1Cz(:,curk,:,:,:,:) = squeeze((P.C(:,nextk,:,:,:,:) - P.C(:,prevk,:,:,:,:))/(2*P.dz));
end
if P.Ela == 0
	d1Nx(curi,:,:,:,:,:) = squeeze((P.N(nexti,:,:,:,:,:) - P.N(previ,:,:,:,:,:))/(2*P.dx));
	d1Nz(:,curk,:,:,:,:) = squeeze((P.N(:,nextk,:,:,:,:) - P.N(:,prevk,:,:,:,:))/(2*P.dz));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETDERIVATIVES 2D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetDerivatives2D(P)

global ux1 uy1 uz1 ...
       ux2 uy2 uz2 ...
       ux3 uy3 uz3 ...
       previ curi nexti ...
       prevk curk nextk ...
       d1uxx1 d1uxz1 ...
       d1uyx1 d1uyz1 ...
       d1uzx1 d1uzz1 ...
       d2uxxx1 d2uxzx1 ...
       d2uyxx1 d2uyzx1 ...
       d2uzxx1 d2uzzx1 ...
       d2uxxz1 d2uxzz1 ...
       d2uyxz1 d2uyzz1 ...
       d2uzxz1 d2uzzz1 ...
       d1uxx2 d1uxz2 ...
       d1uyx2 d1uyz2 ...
       d1uzx2 d1uzz2 ...
       d2uxxx2 d2uxzx2 ...
       d2uyxx2 d2uyzx2 ...
       d2uzxx2 d2uzzx2 ...
       d2uxxz2 d2uxzz2 ...
       d2uyxz2 d2uyzz2 ...
       d2uzxz2 d2uzzz2 ...
       d2uxxt d2uxzt ...
       d2uyxt d2uyzt ...
       d2uzxt d2uzzt ...
       d3uxxxt d3uxzxt ...
       d3uyxxt d3uyzxt ...
       d3uzxxt d3uzzxt ...
       d3uxxzt d3uxzzt ...
       d3uyxzt d3uyzzt ...
       d3uzxzt d3uzzzt;

d1uxx1(curi,:) = (ux1(nexti,:) - ux1(previ,:))/(2*P.dx);
d1uyx1(curi,:) = (uy1(nexti,:) - uy1(previ,:))/(2*P.dx);
d1uzx1(curi,:) = (uz1(nexti,:) - uz1(previ,:))/(2*P.dx);
d1uxz1(:,curk) = (ux1(:,nextk) - ux1(:,prevk))/(2*P.dz);
d1uyz1(:,curk) = (uy1(:,nextk) - uy1(:,prevk))/(2*P.dz);
d1uzz1(:,curk) = (uz1(:,nextk) - uz1(:,prevk))/(2*P.dz);

d2uxxx1(curi,:) = (ux1(nexti,:) - 2*ux1(curi,:) + ux1(previ,:))/(P.dx^2);
d2uyxx1(curi,:) = (uy1(nexti,:) - 2*uy1(curi,:) + uy1(previ,:))/(P.dx^2);
d2uzxx1(curi,:) = (uz1(nexti,:) - 2*uz1(curi,:) + uz1(previ,:))/(P.dx^2);
d2uxxz1(curi,curk) = (ux1(nexti,nextk) - ux1(nexti,prevk) - ux1(previ,nextk) + ux1(previ,prevk))/(4*P.dx*P.dz);
d2uyxz1(curi,curk) = (uy1(nexti,nextk) - uy1(nexti,prevk) - uy1(previ,nextk) + uy1(previ,prevk))/(4*P.dx*P.dz);
d2uzxz1(curi,curk) = (uz1(nexti,nextk) - uz1(nexti,prevk) - uz1(previ,nextk) + uz1(previ,prevk))/(4*P.dx*P.dz);
d2uxzz1(:,curk) = (ux1(:,nextk) - 2*ux1(:,curk) + ux1(:,prevk))/(P.dz^2);
d2uyzz1(:,curk) = (uy1(:,nextk) - 2*uy1(:,curk) + uy1(:,prevk))/(P.dz^2);
d2uzzz1(:,curk) = (uz1(:,nextk) - 2*uz1(:,curk) + uz1(:,prevk))/(P.dz^2);
d2uxzx1(curi,curk) = (ux1(nexti,nextk) - ux1(previ,nextk) - ux1(nexti,prevk) + ux1(previ,prevk))/(4*P.dx*P.dz);
d2uyzx1(curi,curk) = (uy1(nexti,nextk) - uy1(previ,nextk) - uy1(nexti,prevk) + uy1(previ,prevk))/(4*P.dx*P.dz);
d2uzzx1(curi,curk) = (uz1(nexti,nextk) - uz1(previ,nextk) - uz1(nexti,prevk) + uz1(previ,prevk))/(4*P.dx*P.dz);

d1uxx2(curi,:) = (ux2(nexti,:) - ux2(previ,:))/(2*P.dx);
d1uyx2(curi,:) = (uy2(nexti,:) - uy2(previ,:))/(2*P.dx);
d1uzx2(curi,:) = (uz2(nexti,:) - uz2(previ,:))/(2*P.dx);
d1uxz2(:,curk) = (ux2(:,nextk) - ux2(:,prevk))/(2*P.dz);
d1uyz2(:,curk) = (uy2(:,nextk) - uy2(:,prevk))/(2*P.dz);
d1uzz2(:,curk) = (uz2(:,nextk) - uz2(:,prevk))/(2*P.dz);

d2uxxx2(curi,:) = (ux2(nexti,:) - 2*ux2(curi,:) + ux2(previ,:))/(P.dx^2);
d2uyxx2(curi,:) = (uy2(nexti,:) - 2*uy2(curi,:) + uy2(previ,:))/(P.dx^2);
d2uzxx2(curi,:) = (uz2(nexti,:) - 2*uz2(curi,:) + uz2(previ,:))/(P.dx^2);
d2uxxz2(curi,curk) = (ux2(nexti,nextk) - ux2(nexti,prevk) - ux2(previ,nextk) + ux2(previ,prevk))/(4*P.dx*P.dz);
d2uyxz2(curi,curk) = (uy2(nexti,nextk) - uy2(nexti,prevk) - uy2(previ,nextk) + uy2(previ,prevk))/(4*P.dx*P.dz);
d2uzxz2(curi,curk) = (uz2(nexti,nextk) - uz2(nexti,prevk) - uz2(previ,nextk) + uz2(previ,prevk))/(4*P.dx*P.dz);
d2uxzz2(:,curk) = (ux2(:,nextk) - 2*ux2(:,curk) + ux2(:,prevk))/(P.dz^2);
d2uyzz2(:,curk) = (uy2(:,nextk) - 2*uy2(:,curk) + uy2(:,prevk))/(P.dz^2);
d2uzzz2(:,curk) = (uz2(:,nextk) - 2*uz2(:,curk) + uz2(:,prevk))/(P.dz^2);
d2uxzx2(curi,curk) = (ux2(nexti,nextk) - ux2(previ,nextk) - ux2(nexti,prevk) + ux2(previ,prevk))/(4*P.dx*P.dz);
d2uyzx2(curi,curk) = (uy2(nexti,nextk) - uy2(previ,nextk) - uy2(nexti,prevk) + uy2(previ,prevk))/(4*P.dx*P.dz);
d2uzzx2(curi,curk) = (uz2(nexti,nextk) - uz2(previ,nextk) - uz2(nexti,prevk) + uz2(previ,prevk))/(4*P.dx*P.dz);

d2uxxt = (d1uxx2 - d1uxx1)/P.dt;
d2uyxt = (d1uyx2 - d1uyx1)/P.dt;
d2uzxt = (d1uzx2 - d1uzx1)/P.dt;
d2uxzt = (d1uxz2 - d1uxz1)/P.dt;
d2uyzt = (d1uyz2 - d1uyz1)/P.dt;
d2uzzt = (d1uzz2 - d1uzz1)/P.dt;

d3uxxxt = (d2uxxx2 - d2uxxx1)/P.dt;
d3uyxxt = (d2uyxx2 - d2uyxx1)/P.dt;
d3uzxxt = (d2uzxx2 - d2uzxx1)/P.dt;
d3uxxzt = (d2uxxz2 - d2uxxz1)/P.dt;
d3uyxzt = (d2uyxz2 - d2uyxz1)/P.dt;
d3uzxzt = (d2uzxz2 - d2uzxz1)/P.dt;
d3uxzzt = (d2uxzz2 - d2uxzz1)/P.dt;
d3uyzzt = (d2uyzz2 - d2uyzz1)/P.dt;
d3uzzzt = (d2uzzz2 - d2uzzz1)/P.dt;
d3uxzxt = (d2uxzx2 - d2uxzx1)/P.dt;
d3uyzxt = (d2uyzx2 - d2uyzx1)/P.dt;
d3uzzxt = (d2uzzx2 - d2uzzx1)/P.dt;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		APPLYBC 2D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ApplyBC2D(t,P)

global ux2 ux3 ...
       uy2 uy3 ...
       uz2 uz3 ...
       BCyz BCxy ...
       qx2 qt2 qxt2 ...
       qx qt qxt;
      
switch P.BCxy1
	case 0
		ux3(:,1) = BCxy;
		uy3(:,1) = BCxy;
		uz3(:,1) = BCxy;
	case 1
		ux3(:,1) = ux3(:,2);
		uy3(:,1) = uy3(:,2);
		uz3(:,1) = uz3(:,2);
	case 2
		ux3(:,1) = -qx*ux3(:,2) + -qt*ux2(:,1) + -qxt*ux2(:,2);
		uy3(:,1) = -qx*uy3(:,2) + -qt*uy2(:,1) + -qxt*uy2(:,2);
		uz3(:,1) = -qx*uz3(:,2) + -qt*uz2(:,1) + -qxt*uz2(:,2);
end
switch P.BCxy2
	case 0
		ux3(:,P.nz) = BCxy;
		uy3(:,P.nz) = BCxy;
		uz3(:,P.nz) = BCxy;
	case 1
		ux3(:,P.nz) = ux3(:,P.nz-1);
		uy3(:,P.nz) = uy3(:,P.nz-1);
		uz3(:,P.nz) = uz3(:,P.nz-1);
	case 2
		ux3(:,P.nz) = -qx*ux3(:,P.nz-1) + -qt*ux2(:,P.nz) + -qxt*ux2(:,P.nz-1);
		uy3(:,P.nz) = -qx*uy3(:,P.nz-1) + -qt*uy2(:,P.nz) + -qxt*uy2(:,P.nz-1);
		uz3(:,P.nz) = -qx*uz3(:,P.nz-1) + -qt*uz2(:,P.nz) + -qxt*uz2(:,P.nz-1);
end
switch P.BCyz1
	case 0
		ux3(1,:) = BCyz;
		uy3(1,:) = BCyz;
		uz3(1,:) = BCyz;
	case 1
		ux3(1,:) = ux3(2,:);
		uy3(1,:) = uy3(2,:);
		uz3(1,:) = uz3(2,:);
	case 2
		ux3(1,:) = -qx2*ux3(2,:) + -qt2*ux2(1,:) + -qxt2*ux2(2,:);
		uy3(1,:) = -qx2*uy3(2,:) + -qt2*uy2(1,:) + -qxt2*uy2(2,:);
		uz3(1,:) = -qx2*uz3(2,:) + -qt2*uz2(1,:) + -qxt2*uz2(2,:);
end
switch P.BCyz2
	case 0
		ux3(P.nx,:) = BCyz;
		uy3(P.nx,:) = BCyz;
		uz3(P.nx,:) = BCyz;
	case 1
		ux3(P.nx,:) = ux3(P.nx-1,:);
		uy3(P.nx,:) = uy3(P.nx-1,:);
		uz3(P.nx,:) = uz3(P.nx-1,:);
	case 2
		ux3(P.nx,:) = -qx2*ux3(P.nx-1,:) + -qt2*ux2(P.nx,:) + -qxt2*ux2(P.nx-1,:);
		uy3(P.nx,:) = -qx2*uy3(P.nx-1,:) + -qt2*uy2(P.nx,:) + -qxt2*uy2(P.nx-1,:);
		uz3(P.nx,:) = -qx2*uz3(P.nx-1,:) + -qt2*uz2(P.nx,:) + -qxt2*uz2(P.nx-1,:);
end

if t < P.ITnt + 1
	%ux3(P.ITx,P.ITz) = ux3(P.ITx,P.ITz) + P.dITAx(:,:,t);
	%uy3(P.ITx,P.ITz) = uy3(P.ITx,P.ITz) + P.dITAy(:,:,t);
	%uz3(P.ITx,P.ITz) = uz3(P.ITx,P.ITz) + P.dITAz(:,:,t);
	ux3(P.ITx,P.ITz) = P.ITAx(:,:,t);
	uy3(P.ITx,P.ITz) = P.ITAy(:,:,t);
	uz3(P.ITx,P.ITz) = P.ITAz(:,:,t);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		INITIALIZE 3D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function Initialize3D(P)

global u f ...
       BCxy BCxz BCyz ...
       qx qt qxt ...
       qx2 qt2 qxt2 ...
       previ curi nexti ...
       prevj curj nextj ...
       prevk curk nextk ...
       ux1 ux2 ux3 ...
       uy1 uy2 uy3 ...
       uz1 uz2 uz3 ...
       d1uxx1 d1uxy1 d1uxz1 ...
       d1uyx1 d1uyy1 d1uyz1 ...
       d1uzx1 d1uzy1 d1uzz1 ...
       d2uxxx1 d2uxyx1 d2uxzx1 ...
       d2uyxx1 d2uyyx1 d2uyzx1 ...
       d2uzxx1 d2uzyx1 d2uzzx1 ...
       d2uxxy1 d2uxyy1 d2uxzy1 ...
       d2uyxy1 d2uyyy1 d2uyzy1 ...
       d2uzxy1 d2uzyy1 d2uzzy1 ...
       d2uxxz1 d2uxyz1 d2uxzz1 ...
       d2uyxz1 d2uyyz1 d2uyzz1 ...
       d2uzxz1 d2uzyz1 d2uzzz1 ...
       d1uxx2 d1uxy2 d1uxz2 ...
       d1uyx2 d1uyy2 d1uyz2 ...
       d1uzx2 d1uzy2 d1uzz2 ...
       d2uxxx2 d2uxyx2 d2uxzx2 ...
       d2uyxx2 d2uyyx2 d2uyzx2 ...
       d2uzxx2 d2uzyx2 d2uzzx2 ...
       d2uxxy2 d2uxyy2 d2uxzy2 ...
       d2uyxy2 d2uyyy2 d2uyzy2 ...
       d2uzxy2 d2uzyy2 d2uzzy2 ...
       d2uxxz2 d2uxyz2 d2uxzz2 ...
       d2uyxz2 d2uyyz2 d2uyzz2 ...
       d2uzxz2 d2uzyz2 d2uzzz2 ...
       d2uxxt d2uxyt d2uxzt ...
       d2uyxt d2uyyt d2uyzt ...
       d2uzxt d2uzyt d2uzzt ...
       d3uxxxt d3uxyxt d3uxzxt ...
       d3uyxxt d3uyyxt d3uyzxt ...
       d3uzxxt d3uzyxt d3uzzxt ...
       d3uxxyt d3uxyyt d3uxzyt ...
       d3uyxyt d3uyyyt d3uyzyt ...
       d3uzxyt d3uzyyt d3uzzyt ...
       d3uxxzt d3uxyzt d3uxzzt ...
       d3uyxzt d3uyyzt d3uyzzt ...
       d3uzxzt d3uzyzt d3uzzzt ...
       d1Lx d1Ly d1Lz ...
       d1Ux d1Uy d1Uz ...
       d1Cx d1Cy d1Cz ...
       d1Nx d1Ny d1Nz;
      
%Derivatives of material properties

if P.Iso == 1
	d1Lx = zeros(P.nx,P.ny,P.nz);
	d1Ly = zeros(P.nx,P.ny,P.nz);
	d1Lz = zeros(P.nx,P.ny,P.nz);
	d1Ux = zeros(P.nx,P.ny,P.nz);
	d1Uy = zeros(P.nx,P.ny,P.nz);
	d1Uz = zeros(P.nx,P.ny,P.nz);
else
	d1Cx = zeros(P.nx,P.ny,P.nz,3,3,3,3);
	d1Cy = zeros(P.nx,P.ny,P.nz,3,3,3,3);
	d1Cz = zeros(P.nx,P.ny,P.nz,3,3,3,3);
end
if P.Ela == 0
	d1Nx = zeros(P.nx,P.ny,P.nz,3,3,3,3);
	d1Ny = zeros(P.nx,P.ny,P.nz,3,3,3,3);
	d1Nz = zeros(P.nx,P.ny,P.nz,3,3,3,3);
end

%Displacements for three time steps

ux1 = zeros(P.nx,P.ny,P.nz); %Displacement in the x direction for t - 1
ux2 = zeros(P.nx,P.ny,P.nz); %Displacement in the x direction for t 
ux3 = zeros(P.nx,P.ny,P.nz); %Displacement in the x direction for t + 1

uy1 = zeros(P.nx,P.ny,P.nz); %Displacement in the y direction for t - 1
uy2 = zeros(P.nx,P.ny,P.nz); %Displacement in the y direction for t 
uy3 = zeros(P.nx,P.ny,P.nz); %Displacement in the y direction for t + 1

uz1 = zeros(P.nx,P.ny,P.nz); %Displacement in the z direction for t - 1
uz2 = zeros(P.nx,P.ny,P.nz); %Displacement in the z direction for t 
uz3 = zeros(P.nx,P.ny,P.nz); %Displacement in the z direction for t + 1

%Derivatives of displacements

d1uxx1 = zeros(P.nx,P.ny,P.nz); 
d1uxy1 = zeros(P.nx,P.ny,P.nz); 
d1uxz1 = zeros(P.nx,P.ny,P.nz); 
d1uyx1 = zeros(P.nx,P.ny,P.nz); 
d1uyy1 = zeros(P.nx,P.ny,P.nz); 
d1uyz1 = zeros(P.nx,P.ny,P.nz); 
d1uzx1 = zeros(P.nx,P.ny,P.nz); 
d1uzy1 = zeros(P.nx,P.ny,P.nz); 
d1uzz1 = zeros(P.nx,P.ny,P.nz); 

d2uxxx1 = zeros(P.nx,P.ny,P.nz); 
d2uxyx1 = zeros(P.nx,P.ny,P.nz); 
d2uxzx1 = zeros(P.nx,P.ny,P.nz); 
d2uyxx1 = zeros(P.nx,P.ny,P.nz); 
d2uyyx1 = zeros(P.nx,P.ny,P.nz); 
d2uyzx1 = zeros(P.nx,P.ny,P.nz); 
d2uzxx1 = zeros(P.nx,P.ny,P.nz); 
d2uzyx1 = zeros(P.nx,P.ny,P.nz); 
d2uzzx1 = zeros(P.nx,P.ny,P.nz); 
d2uxxy1 = zeros(P.nx,P.ny,P.nz); 
d2uxyy1 = zeros(P.nx,P.ny,P.nz); 
d2uxzy1 = zeros(P.nx,P.ny,P.nz); 
d2uyxy1 = zeros(P.nx,P.ny,P.nz); 
d2uyyy1 = zeros(P.nx,P.ny,P.nz); 
d2uyzy1 = zeros(P.nx,P.ny,P.nz); 
d2uzxy1 = zeros(P.nx,P.ny,P.nz); 
d2uzyy1 = zeros(P.nx,P.ny,P.nz); 
d2uzzy1 = zeros(P.nx,P.ny,P.nz); 
d2uxxz1 = zeros(P.nx,P.ny,P.nz); 
d2uxyz1 = zeros(P.nx,P.ny,P.nz); 
d2uxzz1 = zeros(P.nx,P.ny,P.nz); 
d2uyxz1 = zeros(P.nx,P.ny,P.nz); 
d2uyyz1 = zeros(P.nx,P.ny,P.nz); 
d2uyzz1 = zeros(P.nx,P.ny,P.nz); 
d2uzxz1 = zeros(P.nx,P.ny,P.nz); 
d2uzyz1 = zeros(P.nx,P.ny,P.nz); 
d2uzzz1 = zeros(P.nx,P.ny,P.nz); 

d1uxx2 = zeros(P.nx,P.ny,P.nz); 
d1uxy2 = zeros(P.nx,P.ny,P.nz); 
d1uxz2 = zeros(P.nx,P.ny,P.nz); 
d1uyx2 = zeros(P.nx,P.ny,P.nz); 
d1uyy2 = zeros(P.nx,P.ny,P.nz); 
d1uyz2 = zeros(P.nx,P.ny,P.nz); 
d1uzx2 = zeros(P.nx,P.ny,P.nz); 
d1uzy2 = zeros(P.nx,P.ny,P.nz); 
d1uzz2 = zeros(P.nx,P.ny,P.nz); 

d2uxxx2 = zeros(P.nx,P.ny,P.nz); 
d2uxyx2 = zeros(P.nx,P.ny,P.nz); 
d2uxzx2 = zeros(P.nx,P.ny,P.nz); 
d2uyxx2 = zeros(P.nx,P.ny,P.nz); 
d2uyyx2 = zeros(P.nx,P.ny,P.nz); 
d2uyzx2 = zeros(P.nx,P.ny,P.nz); 
d2uzxx2 = zeros(P.nx,P.ny,P.nz); 
d2uzyx2 = zeros(P.nx,P.ny,P.nz); 
d2uzzx2 = zeros(P.nx,P.ny,P.nz); 
d2uxxy2 = zeros(P.nx,P.ny,P.nz); 
d2uxyy2 = zeros(P.nx,P.ny,P.nz); 
d2uxzy2 = zeros(P.nx,P.ny,P.nz); 
d2uyxy2 = zeros(P.nx,P.ny,P.nz); 
d2uyyy2 = zeros(P.nx,P.ny,P.nz); 
d2uyzy2 = zeros(P.nx,P.ny,P.nz); 
d2uzxy2 = zeros(P.nx,P.ny,P.nz); 
d2uzyy2 = zeros(P.nx,P.ny,P.nz); 
d2uzzy2 = zeros(P.nx,P.ny,P.nz); 
d2uxxz2 = zeros(P.nx,P.ny,P.nz); 
d2uxyz2 = zeros(P.nx,P.ny,P.nz); 
d2uxzz2 = zeros(P.nx,P.ny,P.nz); 
d2uyxz2 = zeros(P.nx,P.ny,P.nz); 
d2uyyz2 = zeros(P.nx,P.ny,P.nz); 
d2uyzz2 = zeros(P.nx,P.ny,P.nz); 
d2uzxz2 = zeros(P.nx,P.ny,P.nz); 
d2uzyz2 = zeros(P.nx,P.ny,P.nz); 
d2uzzz2 = zeros(P.nx,P.ny,P.nz); 

d2uxxt = zeros(P.nx,P.ny,P.nz); 
d2uxyt = zeros(P.nx,P.ny,P.nz); 
d2uxzt = zeros(P.nx,P.ny,P.nz); 
d2uyxt = zeros(P.nx,P.ny,P.nz); 
d2uyyt = zeros(P.nx,P.ny,P.nz); 
d2uyzt = zeros(P.nx,P.ny,P.nz); 
d2uzxt = zeros(P.nx,P.ny,P.nz); 
d2uzyt = zeros(P.nx,P.ny,P.nz); 
d2uzzt = zeros(P.nx,P.ny,P.nz); 

d3uxxxt = zeros(P.nx,P.ny,P.nz); 
d3uxyxt = zeros(P.nx,P.ny,P.nz); 
d3uxzxt = zeros(P.nx,P.ny,P.nz); 
d3uyxxt = zeros(P.nx,P.ny,P.nz); 
d3uyyxt = zeros(P.nx,P.ny,P.nz); 
d3uyzxt = zeros(P.nx,P.ny,P.nz); 
d3uzxxt = zeros(P.nx,P.ny,P.nz); 
d3uzyxt = zeros(P.nx,P.ny,P.nz); 
d3uzzxt = zeros(P.nx,P.ny,P.nz); 
d3uxxyt = zeros(P.nx,P.ny,P.nz); 
d3uxyyt = zeros(P.nx,P.ny,P.nz); 
d3uxzyt = zeros(P.nx,P.ny,P.nz); 
d3uyxyt = zeros(P.nx,P.ny,P.nz); 
d3uyyyt = zeros(P.nx,P.ny,P.nz); 
d3uyzyt = zeros(P.nx,P.ny,P.nz); 
d3uzxyt = zeros(P.nx,P.ny,P.nz); 
d3uzyyt = zeros(P.nx,P.ny,P.nz); 
d3uzzyt = zeros(P.nx,P.ny,P.nz); 
d3uxxzt = zeros(P.nx,P.ny,P.nz); 
d3uxyzt = zeros(P.nx,P.ny,P.nz); 
d3uxzzt = zeros(P.nx,P.ny,P.nz); 
d3uyxzt = zeros(P.nx,P.ny,P.nz); 
d3uyyzt = zeros(P.nx,P.ny,P.nz); 
d3uyzzt = zeros(P.nx,P.ny,P.nz); 
d3uzxzt = zeros(P.nx,P.ny,P.nz); 
d3uzyzt = zeros(P.nx,P.ny,P.nz); 
d3uzzzt = zeros(P.nx,P.ny,P.nz); 

%Free surface boundary condition, du = 0
%Fixed surface boundary condition, u = 0

BCyz = zeros(1,P.ny,P.nz);
BCxz = zeros(P.nx,1,P.nz);
BCxy = zeros(P.nx,P.ny,1);

%Absorbing boundary condition from Higdon, 1991
%Following are absorbing parameters.

B = 1;
v = P.AbVel*P.dt/P.dz;
b = .4;
qx = (b*(B + v) - v)/((B + v)*(1 - b));
qt = (b*(B + v) - B)/((B + v)*(1 - b));
qxt = b/(b-1);

B = 1;
v = P.AbVel2*P.dt/P.dx;
b = .4;
qx2 = (b*(B + v) - v)/((B + v)*(1 - b));
qt2 = (b*(B + v) - B)/((B + v)*(1 - b));
qxt2 = b/(b-1);

%Initialize other variables

previ = 1:P.nx-2;
curi = 2:P.nx-1;
nexti = 3:P.nx;
prevj = 1:P.ny-2;
curj = 2:P.ny-1;
nextj = 3:P.ny;
prevk = 1:P.nz-2;
curk = 2:P.nz-1;
nextk = 3:P.nz;

%Output parameters

nd = length(P.ud);
if nd > 1
	u.ux = zeros(nd,P.nx,P.nt);
	u.uy = zeros(nd,P.nx,P.nt);
	u.uz = zeros(nd,P.nx,P.nt);
else
	u.ux = zeros(P.nx,P.ny,P.nt);
	u.uy = zeros(P.nx,P.ny,P.nt);
	u.uz = zeros(P.nx,P.ny,P.nt);
end
if P.nf > 0
   f = zeros(3*P.nx,P.nz,P.nf);
else
   f = [];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETMATDERIVATIVES 3D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetMatDerivatives3D(P)

global previ curi nexti ...
       prevj curj nextj ...
       prevk curk nextk ...
       d1Lx d1Ly d1Lz ...
       d1Ux d1Uy d1Uz ...
       d1Cx d1Cy d1Cz ...
       d1Nx d1Ny d1Nz;
      
if P.Iso == 1
	d1Lx(curi,:,:) = (P.L(nexti,:,:) - P.L(previ,:,:))/(2*P.dx);
	d1Ly(:,curj,:) = (P.L(:,nextj,:) - P.L(:,prevj,:))/(2*P.dy);
	d1Lz(:,:,curk) = (P.L(:,:,nextk) - P.L(:,:,prevk))/(2*P.dz);
	d1Ux(curi,:,:) = (P.U(nexti,:,:) - P.U(previ,:,:))/(2*P.dx);
	d1Uy(:,curj,:) = (P.U(:,nextj,:) - P.U(:,prevj,:))/(2*P.dy);
	d1Uz(:,:,curk) = (P.U(:,:,nextk) - P.U(:,:,prevk))/(2*P.dz);
else
	d1Cx(curi,:,:,:,:,:,:) = (P.C(nexti,:,:,:,:,:,:) - P.C(previ,:,:,:,:,:,:))/(2*P.dx);
	d1Cy(:,curj,:,:,:,:,:) = (P.C(:,nextj,:,:,:,:,:) - P.C(:,prevj,:,:,:,:,:))/(2*P.dy);
	d1Cz(:,:,curk,:,:,:,:) = (P.C(:,:,nextk,:,:,:,:) - P.C(:,:,prevk,:,:,:,:))/(2*P.dz);
end
if P.Ela == 0
	d1Nx(curi,:,:,:,:,:,:) = (P.N(nexti,:,:,:,:,:,:) - P.N(previ,:,:,:,:,:,:))/(2*P.dx);
	d1Ny(:,curj,:,:,:,:,:) = (P.N(:,nextj,:,:,:,:,:) - P.N(:,prevj,:,:,:,:,:))/(2*P.dy);
	d1Nz(:,:,curk,:,:,:,:) = (P.N(:,:,nextk,:,:,:,:) - P.N(:,:,prevk,:,:,:,:))/(2*P.dz);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GETDERIVATIVES 3D				%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetDerivatives3D(P)

global ux1 uy1 uz1 ...
       ux2 uy2 uz2 ...
       ux3 uy3 uz3 ...
       previ curi nexti ...
       prevj curj nextj ...
       prevk curk nextk ...
       d1uxx1 d1uxy1 d1uxz1 ...
       d1uyx1 d1uyy1 d1uyz1 ...
       d1uzx1 d1uzy1 d1uzz1 ...
       d2uxxx1 d2uxyx1 d2uxzx1 ...
       d2uyxx1 d2uyyx1 d2uyzx1 ...
       d2uzxx1 d2uzyx1 d2uzzx1 ...
       d2uxxy1 d2uxyy1 d2uxzy1 ...
       d2uyxy1 d2uyyy1 d2uyzy1 ...
       d2uzxy1 d2uzyy1 d2uzzy1 ...
       d2uxxz1 d2uxyz1 d2uxzz1 ...
       d2uyxz1 d2uyyz1 d2uyzz1 ...
       d2uzxz1 d2uzyz1 d2uzzz1 ...
       d1uxx2 d1uxy2 d1uxz2 ...
       d1uyx2 d1uyy2 d1uyz2 ...
       d1uzx2 d1uzy2 d1uzz2 ...
       d2uxxx2 d2uxyx2 d2uxzx2 ...
       d2uyxx2 d2uyyx2 d2uyzx2 ...
       d2uzxx2 d2uzyx2 d2uzzx2 ...
       d2uxxy2 d2uxyy2 d2uxzy2 ...
       d2uyxy2 d2uyyy2 d2uyzy2 ...
       d2uzxy2 d2uzyy2 d2uzzy2 ...
       d2uxxz2 d2uxyz2 d2uxzz2 ...
       d2uyxz2 d2uyyz2 d2uyzz2 ...
       d2uzxz2 d2uzyz2 d2uzzz2 ...
       d2uxxt d2uxyt d2uxzt ...
       d2uyxt d2uyyt d2uyzt ...
       d2uzxt d2uzyt d2uzzt ...
       d3uxxxt d3uxyxt d3uxzxt ...
       d3uyxxt d3uyyxt d3uyzxt ...
       d3uzxxt d3uzyxt d3uzzxt ...
       d3uxxyt d3uxyyt d3uxzyt ...
       d3uyxyt d3uyyyt d3uyzyt ...
       d3uzxyt d3uzyyt d3uzzyt ...
       d3uxxzt d3uxyzt d3uxzzt ...
       d3uyxzt d3uyyzt d3uyzzt ...
       d3uzxzt d3uzyzt d3uzzzt;

d1uxx1(curi,:,:) = (ux1(nexti,:,:) - ux1(previ,:,:))/(2*P.dx);
d1uyx1(curi,:,:) = (uy1(nexti,:,:) - uy1(previ,:,:))/(2*P.dx);
d1uzx1(curi,:,:) = (uz1(nexti,:,:) - uz1(previ,:,:))/(2*P.dx);
d1uxy1(:,curj,:) = (ux1(:,nextj,:) - ux1(:,prevj,:))/(2*P.dy);
d1uyy1(:,curj,:) = (uy1(:,nextj,:) - uy1(:,prevj,:))/(2*P.dy);
d1uzy1(:,curj,:) = (uz1(:,nextj,:) - uz1(:,prevj,:))/(2*P.dy);
d1uxz1(:,:,curk) = (ux1(:,:,nextk) - ux1(:,:,prevk))/(2*P.dz);
d1uyz1(:,:,curk) = (uy1(:,:,nextk) - uy1(:,:,prevk))/(2*P.dz);
d1uzz1(:,:,curk) = (uz1(:,:,nextk) - uz1(:,:,prevk))/(2*P.dz);

d2uxxx2(curi,:,:) = (ux1(nexti,:,:) - 2*ux1(curi,:,:) + ux1(previ,:,:))/(P.dx^2);
d2uyxx2(curi,:,:) = (uy1(nexti,:,:) - 2*uy1(curi,:,:) + uy1(previ,:,:))/(P.dx^2);
d2uzxx2(curi,:,:) = (uz1(nexti,:,:) - 2*uz1(curi,:,:) + uz1(previ,:,:))/(P.dx^2);
d2uxxy2(curi,curj,:) = (ux1(nexti,nextj,:) - ux1(nexti,prevj,:) - ux1(previ,nextj,:) + ux1(previ,prevj,:))/(4*P.dx*P.dy);
d2uyxy2(curi,curj,:) = (uy1(nexti,nextj,:) - uy1(nexti,prevj,:) - uy1(previ,nextj,:) + uy1(previ,prevj,:))/(4*P.dx*P.dy);
d2uzxy2(curi,curj,:) = (uz1(nexti,nextj,:) - uz1(nexti,prevj,:) - uz1(previ,nextj,:) + uz1(previ,prevj,:))/(4*P.dx*P.dy);
d2uxxz2(curi,:,curk) = (ux1(nexti,:,nextk) - ux1(nexti,:,prevk) - ux1(previ,:,nextk) + ux1(previ,:,prevk))/(4*P.dx*P.dz);
d2uyxz2(curi,:,curk) = (uy1(nexti,:,nextk) - uy1(nexti,:,prevk) - uy1(previ,:,nextk) + uy1(previ,:,prevk))/(4*P.dx*P.dz);
d2uzxz2(curi,:,curk) = (uz1(nexti,:,nextk) - uz1(nexti,:,prevk) - uz1(previ,:,nextk) + uz1(previ,:,prevk))/(4*P.dx*P.dz);
d2uyyy2(:,curj,:) = (uy1(:,nextj,:) - 2*uy1(:,curj,:) + uy1(:,prevj,:))/(P.dy^2);
d2uxyy2(:,curj,:) = (ux1(:,nextj,:) - 2*ux1(:,curj,:) + ux1(:,prevj,:))/(P.dy^2);
d2uzyy2(:,curj,:) = (uz1(:,nextj,:) - 2*uz1(:,curj,:) + uz1(:,prevj,:))/(P.dy^2);
d2uxyx2(curi,curj,:) = (ux1(nexti,nextj,:) - ux1(previ,nextj,:) - ux1(nexti,prevj,:) + ux1(previ,prevj,:))/(4*P.dx*P.dy);
d2uyyx2(curi,curj,:) = (uy1(nexti,nextj,:) - uy1(previ,nextj,:) - uy1(nexti,prevj,:) + uy1(previ,prevj,:))/(4*P.dx*P.dy);
d2uzyx2(curi,curj,:) = (uz1(nexti,nextj,:) - uz1(previ,nextj,:) - uz1(nexti,prevj,:) + uz1(previ,prevj,:))/(4*P.dx*P.dy);
d2uxyz2(:,curj,curk) = (ux1(:,nextj,nextk) - ux1(:,nextj,prevk) - ux1(:,prevj,nextk) + ux1(:,prevj,prevk))/(4*P.dy*P.dz);
d2uyyz2(:,curj,curk) = (uy1(:,nextj,nextk) - uy1(:,nextj,prevk) - uy1(:,prevj,nextk) + uy1(:,prevj,prevk))/(4*P.dy*P.dz);
d2uzyz2(:,curj,curk) = (uz1(:,nextj,nextk) - uz1(:,nextj,prevk) - uz1(:,prevj,nextk) + uz1(:,prevj,prevk))/(4*P.dy*P.dz);
d2uzzz2(:,:,curk) = (uz1(:,:,nextk) - 2*uz1(:,:,curk) + uz1(:,:,prevk))/(P.dz^2);
d2uxzz2(:,:,curk) = (ux1(:,:,nextk) - 2*ux1(:,:,curk) + ux1(:,:,prevk))/(P.dz^2);
d2uyzz2(:,:,curk) = (uy1(:,:,nextk) - 2*uy1(:,:,curk) + uy1(:,:,prevk))/(P.dz^2);
d2uxzy2(:,curj,curk) = (ux1(:,nextj,nextk) - ux1(:,prevj,nextk) - ux1(:,nextj,prevk) + ux1(:,prevj,prevk))/(4*P.dz*P.dy);
d2uyzy2(:,curj,curk) = (uy1(:,nextj,nextk) - uy1(:,prevj,nextk) - uy1(:,nextj,prevk) + uy1(:,prevj,prevk))/(4*P.dz*P.dy);
d2uzzy2(:,curj,curk) = (uz1(:,nextj,nextk) - uz1(:,prevj,nextk) - uz1(:,nextj,prevk) + uz1(:,prevj,prevk))/(4*P.dz*P.dy);
d2uxzx2(curi,:,curk) = (ux1(nexti,:,nextk) - ux1(previ,:,nextk) - ux1(nexti,:,prevk) + ux1(previ,:,prevk))/(4*P.dx*P.dz);
d2uyzx2(curi,:,curk) = (uy1(nexti,:,nextk) - uy1(previ,:,nextk) - uy1(nexti,:,prevk) + uy1(previ,:,prevk))/(4*P.dx*P.dz);
d2uzzx2(curi,:,curk) = (uz1(nexti,:,nextk) - uz1(previ,:,nextk) - uz1(nexti,:,prevk) + uz1(previ,:,prevk))/(4*P.dx*P.dz);

d1uxx2(curi,:,:) = (ux2(nexti,:,:) - ux2(previ,:,:))/(2*P.dx);
d1uyx2(curi,:,:) = (uy2(nexti,:,:) - uy2(previ,:,:))/(2*P.dx);
d1uzx2(curi,:,:) = (uz2(nexti,:,:) - uz2(previ,:,:))/(2*P.dx);
d1uxy2(:,curj,:) = (ux2(:,nextj,:) - ux2(:,prevj,:))/(2*P.dy);
d1uyy2(:,curj,:) = (uy2(:,nextj,:) - uy2(:,prevj,:))/(2*P.dy);
d1uzy2(:,curj,:) = (uz2(:,nextj,:) - uz2(:,prevj,:))/(2*P.dy);
d1uxz2(:,:,curk) = (ux2(:,:,nextk) - ux2(:,:,prevk))/(2*P.dz);
d1uyz2(:,:,curk) = (uy2(:,:,nextk) - uy2(:,:,prevk))/(2*P.dz);
d1uzz2(:,:,curk) = (uz2(:,:,nextk) - uz2(:,:,prevk))/(2*P.dz);

d2uxxx2(curi,:,:) = (ux2(nexti,:,:) - 2*ux2(curi,:,:) + ux2(previ,:,:))/(P.dx^2);
d2uyxx2(curi,:,:) = (uy2(nexti,:,:) - 2*uy2(curi,:,:) + uy2(previ,:,:))/(P.dx^2);
d2uzxx2(curi,:,:) = (uz2(nexti,:,:) - 2*uz2(curi,:,:) + uz2(previ,:,:))/(P.dx^2);
d2uxxy2(curi,curj,:) = (ux2(nexti,nextj,:) - ux2(nexti,prevj,:) - ux2(previ,nextj,:) + ux2(previ,prevj,:))/(4*P.dx*P.dy);
d2uyxy2(curi,curj,:) = (uy2(nexti,nextj,:) - uy2(nexti,prevj,:) - uy2(previ,nextj,:) + uy2(previ,prevj,:))/(4*P.dx*P.dy);
d2uzxy2(curi,curj,:) = (uz2(nexti,nextj,:) - uz2(nexti,prevj,:) - uz2(previ,nextj,:) + uz2(previ,prevj,:))/(4*P.dx*P.dy);
d2uxxz2(curi,:,curk) = (ux2(nexti,:,nextk) - ux2(nexti,:,prevk) - ux2(previ,:,nextk) + ux2(previ,:,prevk))/(4*P.dx*P.dz);
d2uyxz2(curi,:,curk) = (uy2(nexti,:,nextk) - uy2(nexti,:,prevk) - uy2(previ,:,nextk) + uy2(previ,:,prevk))/(4*P.dx*P.dz);
d2uzxz2(curi,:,curk) = (uz2(nexti,:,nextk) - uz2(nexti,:,prevk) - uz2(previ,:,nextk) + uz2(previ,:,prevk))/(4*P.dx*P.dz);
d2uyyy2(:,curj,:) = (uy2(:,nextj,:) - 2*uy2(:,curj,:) + uy2(:,prevj,:))/(P.dy^2);
d2uxyy2(:,curj,:) = (ux2(:,nextj,:) - 2*ux2(:,curj,:) + ux2(:,prevj,:))/(P.dy^2);
d2uzyy2(:,curj,:) = (uz2(:,nextj,:) - 2*uz2(:,curj,:) + uz2(:,prevj,:))/(P.dy^2);
d2uxyx2(curi,curj,:) = (ux2(nexti,nextj,:) - ux2(previ,nextj,:) - ux2(nexti,prevj,:) + ux2(previ,prevj,:))/(4*P.dx*P.dy);
d2uyyx2(curi,curj,:) = (uy2(nexti,nextj,:) - uy2(previ,nextj,:) - uy2(nexti,prevj,:) + uy2(previ,prevj,:))/(4*P.dx*P.dy);
d2uzyx2(curi,curj,:) = (uz2(nexti,nextj,:) - uz2(previ,nextj,:) - uz2(nexti,prevj,:) + uz2(previ,prevj,:))/(4*P.dx*P.dy);
d2uxyz2(:,curj,curk) = (ux2(:,nextj,nextk) - ux2(:,nextj,prevk) - ux2(:,prevj,nextk) + ux2(:,prevj,prevk))/(4*P.dy*P.dz);
d2uyyz2(:,curj,curk) = (uy2(:,nextj,nextk) - uy2(:,nextj,prevk) - uy2(:,prevj,nextk) + uy2(:,prevj,prevk))/(4*P.dy*P.dz);
d2uzyz2(:,curj,curk) = (uz2(:,nextj,nextk) - uz2(:,nextj,prevk) - uz2(:,prevj,nextk) + uz2(:,prevj,prevk))/(4*P.dy*P.dz);
d2uzzz2(:,:,curk) = (uz2(:,:,nextk) - 2*uz2(:,:,curk) + uz2(:,:,prevk))/(P.dz^2);
d2uxzz2(:,:,curk) = (ux2(:,:,nextk) - 2*ux2(:,:,curk) + ux2(:,:,prevk))/(P.dz^2);
d2uyzz2(:,:,curk) = (uy2(:,:,nextk) - 2*uy2(:,:,curk) + uy2(:,:,prevk))/(P.dz^2);
d2uxzy2(:,curj,curk) = (ux2(:,nextj,nextk) - ux2(:,prevj,nextk) - ux2(:,nextj,prevk) + ux2(:,prevj,prevk))/(4*P.dz*P.dy);
d2uyzy2(:,curj,curk) = (uy2(:,nextj,nextk) - uy2(:,prevj,nextk) - uy2(:,nextj,prevk) + uy2(:,prevj,prevk))/(4*P.dz*P.dy);
d2uzzy2(:,curj,curk) = (uz2(:,nextj,nextk) - uz2(:,prevj,nextk) - uz2(:,nextj,prevk) + uz2(:,prevj,prevk))/(4*P.dz*P.dy);
d2uxzx2(curi,:,curk) = (ux2(nexti,:,nextk) - ux2(previ,:,nextk) - ux2(nexti,:,prevk) + ux2(previ,:,prevk))/(4*P.dx*P.dz);
d2uyzx2(curi,:,curk) = (uy2(nexti,:,nextk) - uy2(previ,:,nextk) - uy2(nexti,:,prevk) + uy2(previ,:,prevk))/(4*P.dx*P.dz);
d2uzzx2(curi,:,curk) = (uz2(nexti,:,nextk) - uz2(previ,:,nextk) - uz2(nexti,:,prevk) + uz2(previ,:,prevk))/(4*P.dx*P.dz);

d2uxxt = (d1uxx2 - d1uxx1)/P.dt;
d2uyxt = (d1uyx2 - d1uyx1)/P.dt;
d2uzxt = (d1uzx2 - d1uzx1)/P.dt;
d2uxyt = (d1uxy2 - d1uxy1)/P.dt;
d2uyyt = (d1uyy2 - d1uyy1)/P.dt;
d2uzyt = (d1uzy2 - d1uzy1)/P.dt;
d2uxzt = (d1uxz2 - d1uxz1)/P.dt;
d2uyzt = (d1uyz2 - d1uyz1)/P.dt;
d2uzzt = (d1uzz2 - d1uzz1)/P.dt;

d3uxxxt = (d2uxxx2 - d2uxxx1)/P.dt;
d3uyxxt = (d2uyxx2 - d2uyxx1)/P.dt;
d3uzxxt = (d2uzxx2 - d2uzxx1)/P.dt;
d3uxxyt = (d2uxxy2 - d2uxxy1)/P.dt;
d3uyxyt = (d2uyxy2 - d2uyxy1)/P.dt;
d3uzxyt = (d2uzxy2 - d2uzxy1)/P.dt;
d3uxxzt = (d2uxxz2 - d2uxxz1)/P.dt;
d3uyxzt = (d2uyxz2 - d2uyxz1)/P.dt;
d3uzxzt = (d2uzxz2 - d2uzxz1)/P.dt;
d3uyyyt = (d2uyyy2 - d2uyyy1)/P.dt;
d3uxyyt = (d2uxyy2 - d2uxyy1)/P.dt;
d3uzyyt = (d2uzyy2 - d2uzyy1)/P.dt;
d3uxyxt = (d2uxyx2 - d2uxyx1)/P.dt;
d3uyyxt = (d2uyyx2 - d2uyyx1)/P.dt;
d3uzyxt = (d2uzyx2 - d2uzyx1)/P.dt;
d3uxyzt = (d2uxyz2 - d2uxyz1)/P.dt;
d3uyyzt = (d2uyyz2 - d2uyyz1)/P.dt;
d3uzyzt = (d2uzyz2 - d2uzyz1)/P.dt;
d3uzzzt = (d2uzzz2 - d2uzzz1)/P.dt;
d3uxzzt = (d2uxzz2 - d2uxzz1)/P.dt;
d3uyzzt = (d2uyzz2 - d2uyzz1)/P.dt;
d3uxzyt = (d2uxzy2 - d2uxzy1)/P.dt;
d3uyzyt = (d2uyzy2 - d2uyzy1)/P.dt;
d3uzzyt = (d2uzzy2 - d2uzzy1)/P.dt;
d3uxzxt = (d2uxzx2 - d2uxzx1)/P.dt;
d3uyzxt = (d2uyzx2 - d2uyzx1)/P.dt;
d3uzzxt = (d2uzzx2 - d2uzzx1)/P.dt;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		APPLYBC 3D					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function ApplyBC3D(t,P)

global ux2 ux3 ...
       uy2 uy3 ...
       uz2 uz3 ...
       BCyz BCxz BCxy ...
       qx2 qt2 qxt2 ...
       qx qt qxt;

switch P.BCxy1
	case 0
		ux3(:,:,1) = BCxy;
		uy3(:,:,1) = BCxy;
		uz3(:,:,1) = BCxy;
	case 1
		ux3(:,:,1) = ux3(:,:,2);
		uy3(:,:,1) = uy3(:,:,2);
		uz3(:,:,1) = uz3(:,:,2);
	case 2
		ux3(:,:,1) = -qx*ux3(:,:,2) + -qt*ux2(:,:,1) + -qxt*ux2(:,:,2);
		uy3(:,:,1) = -qx*uy3(:,:,2) + -qt*uy2(:,:,1) + -qxt*uy2(:,:,2);
		uz3(:,:,1) = -qx*uz3(:,:,2) + -qt*uz2(:,:,1) + -qxt*uz2(:,:,2);
end
switch P.BCxy2
	case 0
		ux3(:,:,P.nz) = BCxy;
		uy3(:,:,P.nz) = BCxy;
		uz3(:,:,P.nz) = BCxy;
	case 1
		ux3(:,:,P.nz) = ux3(:,:,P.nz-1);
		uy3(:,:,P.nz) = uy3(:,:,P.nz-1);
		uz3(:,:,P.nz) = uz3(:,:,P.nz-1);
	case 2
		ux3(:,:,P.nz) = -qx*ux3(:,:,P.nz-1) + -qt*ux2(:,:,P.nz) + -qxt*ux2(:,:,P.nz-1);
		uy3(:,:,P.nz) = -qx*uy3(:,:,P.nz-1) + -qt*uy2(:,:,P.nz) + -qxt*uy2(:,:,P.nz-1);
		uz3(:,:,P.nz) = -qx*uz3(:,:,P.nz-1) + -qt*uz2(:,:,P.nz) + -qxt*uz2(:,:,P.nz-1);
end
switch P.BCxz1
	case 0
		ux3(:,1,:) = BCxz;
		uy3(:,1,:) = BCxz;
		uz3(:,1,:) = BCxz;
	case 1
		ux3(:,1,:) = ux3(:,2,:);
		uy3(:,1,:) = uy3(:,2,:);
		uz3(:,1,:) = uz3(:,2,:);
	case 2
		ux3(:,1,:) = -qx2*ux3(:,2,:) + -qt2*ux2(:,1,:) + -qxt2*ux2(:,2,:);
		uy3(:,1,:) = -qx2*uy3(:,2,:) + -qt2*uy2(:,1,:) + -qxt2*uy2(:,2,:);
		uz3(:,1,:) = -qx2*uz3(:,2,:) + -qt2*uz2(:,1,:) + -qxt2*uz2(:,2,:);
end
switch P.BCxz2
	case 0
		ux3(:,P.ny,:) = BCxz;
		uy3(:,P.ny,:) = BCxz;
		uz3(:,P.ny,:) = BCxz;
	case 1
		ux3(:,P.ny,:) = ux3(:,P.ny-1,:);
		uy3(:,P.ny,:) = uy3(:,P.ny-1,:);
		uz3(:,P.ny,:) = uz3(:,P.ny-1,:);
	case 2
		ux3(:,P.ny,:) = -qx2*ux3(:,P.ny-1,:) + -qt2*ux2(:,P.ny,:) + -qxt2*ux2(:,P.ny-1,:);
		uy3(:,P.ny,:) = -qx2*uy3(:,P.ny-1,:) + -qt2*uy2(:,P.ny,:) + -qxt2*uy2(:,P.ny-1,:);
		uz3(:,P.ny,:) = -qx2*uz3(:,P.ny-1,:) + -qt2*uz2(:,P.ny,:) + -qxt2*uz2(:,P.ny-1,:);
end
switch P.BCyz1
	case 0
		ux3(1,:,:) = BCyz;
		uy3(1,:,:) = BCyz;
		uz3(1,:,:) = BCyz;
	case 1
		ux3(1,:,:) = ux3(2,:,:);
		uy3(1,:,:) = uy3(2,:,:);
		uz3(1,:,:) = uz3(2,:,:);
	case 2
		ux3(1,:,:) = -qx2*ux3(2,:,:) + -qt2*ux2(1,:,:) + -qxt2*ux2(2,:,:);
		uy3(1,:,:) = -qx2*uy3(2,:,:) + -qt2*uy2(1,:,:) + -qxt2*uy2(2,:,:);
		uz3(1,:,:) = -qx2*uz3(2,:,:) + -qt2*uz2(1,:,:) + -qxt2*uz2(2,:,:);
end
switch P.BCyz2
	case 0
		ux3(P.nx,:,:) = BCyz;
		uy3(P.nx,:,:) = BCyz;
		uz3(P.nx,:,:) = BCyz;
	case 1
		ux3(P.nx,:,:) = ux3(P.nx-1,:,:);
		uy3(P.nx,:,:) = uy3(P.nx-1,:,:);
		uz3(P.nx,:,:) = uz3(P.nx-1,:,:);
	case 2
		ux3(P.nx,:,:) = -qx2*ux3(P.nx-1,:,:) + -qt2*ux2(P.nx,:,:) + -qxt2*ux2(P.nx-1,:,:);
		uy3(P.nx,:,:) = -qx2*uy3(P.nx-1,:,:) + -qt2*uy2(P.nx,:,:) + -qxt2*uy2(P.nx-1,:,:);
		uz3(P.nx,:,:) = -qx2*uz3(P.nx-1,:,:) + -qt2*uz2(P.nx,:,:) + -qxt2*uz2(P.nx-1,:,:);
end

if t < P.ITnt + 1
	%ux3(P.ITx,P.ITy,P.ITz) = ux3(P.ITx,P.ITy,P.ITz) + P.dITAx(:,:,:,t);
	%uy3(P.ITx,P.ITy,P.ITz) = uy3(P.ITx,P.ITy,P.ITz) + P.dITAy(:,:,:,t);
	%uz3(P.ITx,P.ITy,P.ITz) = uz3(P.ITx,P.ITy,P.ITz) + P.dITAz(:,:,:,t);
	ux3(P.ITx,P.ITy,P.ITz) = P.ITAx(:,:,:,t);
	uy3(P.ITx,P.ITy,P.ITz) = P.ITAy(:,:,:,t);
	uz3(P.ITx,P.ITy,P.ITz) = P.ITAz(:,:,:,t);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		RHS Calcs					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [rhsX, rhsY, rhsZ] = rhs(P,t)

global d1uxx1 d1uxy1 d1uxz1 ...
       d1uyx1 d1uyy1 d1uyz1 ...
       d1uzx1 d1uzy1 d1uzz1 ...
       d2uxxx1 d2uxyx1 d2uxzx1 ...
       d2uyxx1 d2uyyx1 d2uyzx1 ...
       d2uzxx1 d2uzyx1 d2uzzx1 ...
       d2uxxy1 d2uxyy1 d2uxzy1 ...
       d2uyxy1 d2uyyy1 d2uyzy1 ...
       d2uzxy1 d2uzyy1 d2uzzy1 ...
       d2uxxz1 d2uxyz1 d2uxzz1 ...
       d2uyxz1 d2uyyz1 d2uyzz1 ...
       d2uzxz1 d2uzyz1 d2uzzz1 ...
       d1uxx2 d1uxy2 d1uxz2 ...
       d1uyx2 d1uyy2 d1uyz2 ...
       d1uzx2 d1uzy2 d1uzz2 ...
       d2uxxx2 d2uxyx2 d2uxzx2 ...
       d2uyxx2 d2uyyx2 d2uyzx2 ...
       d2uzxx2 d2uzyx2 d2uzzx2 ...
       d2uxxy2 d2uxyy2 d2uxzy2 ...
       d2uyxy2 d2uyyy2 d2uyzy2 ...
       d2uzxy2 d2uzyy2 d2uzzy2 ...
       d2uxxz2 d2uxyz2 d2uxzz2 ...
       d2uyxz2 d2uyyz2 d2uyzz2 ...
       d2uzxz2 d2uzyz2 d2uzzz2 ...
       d2uxxt d2uxyt d2uxzt ...
       d2uyxt d2uyyt d2uyzt ...
       d2uzxt d2uzyt d2uzzt ...
       d3uxxxt d3uxyxt d3uxzxt ...
       d3uyxxt d3uyyxt d3uyzxt ...
       d3uzxxt d3uzyxt d3uzzxt ...
       d3uxxyt d3uxyyt d3uxzyt ...
       d3uyxyt d3uyyyt d3uyzyt ...
       d3uzxyt d3uzyyt d3uzzyt ...
       d3uxxzt d3uxyzt d3uxzzt ...
       d3uyxzt d3uyyzt d3uyzzt ...
       d3uzxzt d3uzyzt d3uzzzt ...
       d1Lx d1Ly d1Lz ...
       d1Ux d1Uy d1Uz ...
       d1Cx d1Cy d1Cz ...
       d1Nx d1Ny d1Nz;

switch t
   case 1;	%	1D, Isotropic
      rhsX = d1Uz.*d1uxz2 + P.U.*d2uxzz2;
      rhsY = d1Uz.*d1uyz2 + P.U.*d2uyzz2;
      rhsZ = d1Lz.*d1uzz2 + P.L.*d2uzzz2 + ...
             d1Uz.*(d1uzz2 + d1uzz2) + P.U.*(d2uzzz2 + d2uzzz2);
   case 2;	%	2D, Isotropic
      trd1u = d1uxx2 + d1uzz2;
      rhsX = d1Lx.*trd1u + P.L.*(d2uxxx2 + d2uzzx2) + ...
             d1Ux.*(d1uxx2 + d1uxx2) + d1Uz.*(d1uzx2 + d1uxz2) + ...
             P.U.*(d2uxxx2 + d2uxxx2 + d2uzxz2 + d2uxzz2);
      rhsY = d1Ux.*d1uyx2 + d1Uz.*d1uyz2 + P.U.*(d2uyxx2 + d2uyzz2);
      rhsZ = d1Lz.*trd1u + P.L.*(d2uxxz2 + d2uzzz2) + ...
             d1Ux.*(d1uxz2 + d1uzx2) + d1Uz.*(d1uzz2 + d1uzz2) + ...
             P.U.*(d2uxzx2 + d2uzxx2 + d2uzzz2 + d2uzzz2);
   case 3;	%	3D, Isotropic
      trd1u = d1uxx2 + d1uyy2 + d1uzz2;
      rhsX = d1Lx.*trd1u + P.L.*(d2uxxx2 + d2uyyx2 + d2uzzx2) + ...
             d1Ux.*(d1uxx2 + d1uxx2) + d1Uy.*(d1uyx2 + d1uxy2) + d1Uz.*(d1uzx2 + d1uxz2) + ...
             P.U.*(d2uxxx2 + d2uxxx2 + d2uyxy2 + d2uxyy2 + d2uzxz2 + d2uxzz2);
      rhsY = d1Ly.*trd1u + P.L.*(d2uxxy2 + d2uyyy2 + d2uzzy2) + ...
             d1Ux.*(d1uxy2 + d1uyx2) + d1Uy.*(d1uyy2 + d1uyy2) + d1Uz.*(d1uzy2 + d1uyz2) + ...
             P.U.*(d2uxyx2 + d2uyxx2 + d2uyyy2 + d2uyyy2 + d2uzyz2 + d2uyzz2);
      rhsZ = d1Lz.*trd1u + P.L.*(d2uxxz2 + d2uyyz2 + d2uzzz2) + ...
             d1Ux.*(d1uxz2 + d1uzx2) + d1Uy.*(d1uyz2 + d1uzy2) + d1Uz.*(d1uzz2 + d1uzz2) + ...
             P.U.*(d2uxzx2 + d2uzxx2 + d2uyzy2 + d2uzyy2 + d2uzzz2 + d2uzzz2);
   case 4;	%	1D, Anisotropic
      rhsX = P.C(:,1,3,3,3)'.*d2uzzz2 + d1Cz(:,1,3,3,3)'.*d1uzz2 + ...
             P.C(:,1,3,1,3)'.*d2uxzz2 + d1Cz(:,1,3,1,3)'.*d1uxz2 + ...
             P.C(:,1,3,2,3)'.*d2uyzz2 + d1Cz(:,1,3,2,3)'.*d1uyz2;
      rhsY = P.C(:,2,3,3,3)'.*d2uzzz2 + d1Cz(:,2,3,3,3)'.*d1uzz2 + ...
             P.C(:,2,3,1,3)'.*d2uxzz2 + d1Cz(:,2,3,1,3)'.*d1uxz2 + ...
             P.C(:,2,3,2,3)'.*d2uyzz2 + d1Cz(:,2,3,2,3)'.*d1uyz2;
      rhsZ = P.C(:,3,3,3,3)'.*d2uzzz2 + d1Cz(:,3,3,3,3)'.*d1uzz2 + ...
             P.C(:,3,3,1,3)'.*d2uxzz2 + d1Cz(:,3,3,1,3)'.*d1uxz2 + ...
             P.C(:,3,3,2,3)'.*d2uyzz2 + d1Cz(:,3,3,2,3)'.*d1uyz2;
   case 5;	%	2D, Anisotropic
      rhsX = P.C(:,:,1,1,1,1).*d2uxxx2 + d1Cx(:,:,1,1,1,1).*d1uxx2 + ...
             P.C(:,:,1,1,3,3).*d2uzxz2 + d1Cx(:,:,1,1,3,3).*d1uzz2 + ...
             P.C(:,:,1,3,1,1).*d2uxzx2 + d1Cz(:,:,1,3,1,1).*d1uxx2 + ...
             P.C(:,:,1,3,3,3).*d2uzzz2 + d1Cz(:,:,1,3,3,3).*d1uzz2 + ...
             P.C(:,:,1,1,1,2).*(d2uyxx2) +            d1Cx(:,:,1,1,1,2).*(d1uyx2) + ...
             P.C(:,:,1,1,1,3).*(d2uzxx2  + d2uxxz2) + d1Cx(:,:,1,1,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,1,1,2,3).*(d2uyxz2) +            d1Cx(:,:,1,1,2,3).*(d1uyz2) + ...
             P.C(:,:,1,3,1,2).*(d2uyzx2) +            d1Cz(:,:,1,3,1,2).*(d1uyx2) + ...
             P.C(:,:,1,3,1,3).*(d2uzzx2  + d2uxzz2) + d1Cz(:,:,1,3,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,1,3,2,3).*(d2uyzz2) +            d1Cz(:,:,1,3,2,3).*(d1uyz2);
      rhsY = P.C(:,:,2,1,1,1).*d2uxxx2 + d1Cx(:,:,2,1,1,1).*d1uxx2 + ...
             P.C(:,:,2,1,3,3).*d2uzxz2 + d1Cx(:,:,2,1,3,3).*d1uzz2 + ...
             P.C(:,:,2,3,1,1).*d2uxzx2 + d1Cz(:,:,2,3,1,1).*d1uxx2 + ...
             P.C(:,:,2,3,3,3).*d2uzzz2 + d1Cz(:,:,2,3,3,3).*d1uzz2 + ...
             P.C(:,:,2,1,1,2).*(d2uyxx2) +            d1Cx(:,:,2,1,1,2).*(d1uyx2) + ...
             P.C(:,:,2,1,1,3).*(d2uzxx2  + d2uxxz2) + d1Cx(:,:,2,1,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,2,1,2,3).*(d2uyxz2) +            d1Cx(:,:,2,1,2,3).*(d1uyz2) + ...
             P.C(:,:,2,3,1,2).*(d2uyzx2) +            d1Cz(:,:,2,3,1,2).*(d1uyx2) + ...
             P.C(:,:,2,3,1,3).*(d2uzzx2  + d2uxzz2) + d1Cz(:,:,2,3,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,2,3,2,3).*(d2uyzz2) +            d1Cz(:,:,2,3,2,3).*(d1uyz2);
      rhsZ = P.C(:,:,3,1,1,1).*d2uxxx2 + d1Cx(:,:,3,1,1,1).*d1uxx2 + ...
             P.C(:,:,3,1,3,3).*d2uzxz2 + d1Cx(:,:,3,1,3,3).*d1uzz2 + ...
             P.C(:,:,3,3,1,1).*d2uxzx2 + d1Cz(:,:,3,3,1,1).*d1uxx2 + ...
             P.C(:,:,3,3,3,3).*d2uzzz2 + d1Cz(:,:,3,3,3,3).*d1uzz2 + ...
             P.C(:,:,3,1,1,2).*(d2uyxx2) +            d1Cx(:,:,3,1,1,2).*(d1uyx2) + ...
             P.C(:,:,3,1,1,3).*(d2uzxx2  + d2uxxz2) + d1Cx(:,:,3,1,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,3,1,2,3).*(d2uyxz2) +            d1Cx(:,:,3,1,2,3).*(d1uyz2) + ...
             P.C(:,:,3,3,1,2).*(d2uyzx2) +            d1Cz(:,:,3,3,1,2).*(d1uyx2) + ...
             P.C(:,:,3,3,1,3).*(d2uzzx2  + d2uxzz2) + d1Cz(:,:,3,3,1,3).*(d1uzx2  + d1uxz2) + ...
             P.C(:,:,3,3,2,3).*(d2uyzz2) +            d1Cz(:,:,3,3,2,3).*(d1uyz2);
   case 6;	%	3D, Anisotropic
      rhsX = P.C(:,:,:,1,1,1,1).*d2uxxx2 + d1Cx(:,:,:,1,1,1,1).*d1uxx2 + ...
             P.C(:,:,:,1,1,2,2).*d2uyxy2 + d1Cx(:,:,:,1,1,2,2).*d1uyy2 + ...
             P.C(:,:,:,1,1,3,3).*d2uzxz2 + d1Cx(:,:,:,1,1,3,3).*d1uzz2 + ...
             P.C(:,:,:,1,2,1,1).*d2uxyx2 + d1Cy(:,:,:,1,2,1,1).*d1uxx2 + ...
             P.C(:,:,:,1,2,2,2).*d2uyyy2 + d1Cy(:,:,:,1,2,2,2).*d1uyy2 + ...
             P.C(:,:,:,1,2,3,3).*d2uzyz2 + d1Cy(:,:,:,1,2,3,3).*d1uzz2 + ...
             P.C(:,:,:,1,3,1,1).*d2uxzx2 + d1Cz(:,:,:,1,3,1,1).*d1uxx2 + ...
             P.C(:,:,:,1,3,2,2).*d2uyzy2 + d1Cz(:,:,:,1,3,2,2).*d1uyy2 + ...
             P.C(:,:,:,1,3,3,3).*d2uzzz2 + d1Cz(:,:,:,1,3,3,3).*d1uzz2 + ...
             P.C(:,:,:,1,1,1,2).*(d2uyxx2 + d2uxxy2) + d1Cx(:,:,:,1,1,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,1,1,1,3).*(d2uzxx2 + d2uxxz2) + d1Cx(:,:,:,1,1,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,1,1,2,3).*(d2uzxy2 + d2uyxz2) + d1Cx(:,:,:,1,1,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,1,2,1,2).*(d2uyyx2 + d2uxyy2) + d1Cy(:,:,:,1,2,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,1,2,1,3).*(d2uzyx2 + d2uxyz2) + d1Cy(:,:,:,1,2,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,1,2,2,3).*(d2uzyy2 + d2uyyz2) + d1Cy(:,:,:,1,2,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,1,3,1,2).*(d2uyzx2 + d2uxzy2) + d1Cz(:,:,:,1,3,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,1,3,1,3).*(d2uzzx2 + d2uxzz2) + d1Cz(:,:,:,1,3,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,1,3,2,3).*(d2uzzy2 + d2uyzz2) + d1Cz(:,:,:,1,3,2,3).*(d1uzy2 + d1uyz2);
      rhsY = P.C(:,:,:,2,1,1,1).*d2uxxx2 + d1Cx(:,:,:,2,1,1,1).*d1uxx2 + ...
             P.C(:,:,:,2,1,2,2).*d2uyxy2 + d1Cx(:,:,:,2,1,2,2).*d1uyy2 + ...
             P.C(:,:,:,2,1,3,3).*d2uzxz2 + d1Cx(:,:,:,2,1,3,3).*d1uzz2 + ...
             P.C(:,:,:,2,2,1,1).*d2uxyx2 + d1Cy(:,:,:,2,2,1,1).*d1uxx2 + ...
             P.C(:,:,:,2,2,2,2).*d2uyyy2 + d1Cy(:,:,:,2,2,2,2).*d1uyy2 + ...
             P.C(:,:,:,2,2,3,3).*d2uzyz2 + d1Cy(:,:,:,2,2,3,3).*d1uzz2 + ...
             P.C(:,:,:,2,3,1,1).*d2uxzx2 + d1Cz(:,:,:,2,3,1,1).*d1uxx2 + ...
             P.C(:,:,:,2,3,2,2).*d2uyzy2 + d1Cz(:,:,:,2,3,2,2).*d1uyy2 + ...
             P.C(:,:,:,2,3,3,3).*d2uzzz2 + d1Cz(:,:,:,2,3,3,3).*d1uzz2 + ...
             P.C(:,:,:,2,1,1,2).*(d2uyxx2 + d2uxxy2) + d1Cx(:,:,:,2,1,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,2,1,1,3).*(d2uzxx2 + d2uxxz2) + d1Cx(:,:,:,2,1,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,2,1,2,3).*(d2uzxy2 + d2uyxz2) + d1Cx(:,:,:,2,1,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,2,2,1,2).*(d2uyyx2 + d2uxyy2) + d1Cy(:,:,:,2,2,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,2,2,1,3).*(d2uzyx2 + d2uxyz2) + d1Cy(:,:,:,2,2,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,2,2,2,3).*(d2uzyy2 + d2uyyz2) + d1Cy(:,:,:,2,2,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,2,3,1,2).*(d2uyzx2 + d2uxzy2) + d1Cz(:,:,:,2,3,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,2,3,1,3).*(d2uzzx2 + d2uxzz2) + d1Cz(:,:,:,2,3,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,2,3,2,3).*(d2uzzy2 + d2uyzz2) + d1Cz(:,:,:,2,3,2,3).*(d1uzy2 + d1uyz2);
      rhsZ = P.C(:,:,:,3,1,1,1).*d2uxxx2 + d1Cx(:,:,:,3,1,1,1).*d1uxx2 + ...
             P.C(:,:,:,3,1,2,2).*d2uyxy2 + d1Cx(:,:,:,3,1,2,2).*d1uyy2 + ...
             P.C(:,:,:,3,1,3,3).*d2uzxz2 + d1Cx(:,:,:,3,1,3,3).*d1uzz2 + ...
             P.C(:,:,:,3,2,1,1).*d2uxyx2 + d1Cy(:,:,:,3,2,1,1).*d1uxx2 + ...
             P.C(:,:,:,3,2,2,2).*d2uyyy2 + d1Cy(:,:,:,3,2,2,2).*d1uyy2 + ...
             P.C(:,:,:,3,2,3,3).*d2uzyz2 + d1Cy(:,:,:,3,2,3,3).*d1uzz2 + ...
             P.C(:,:,:,3,3,1,1).*d2uxzx2 + d1Cz(:,:,:,3,3,1,1).*d1uxx2 + ...
             P.C(:,:,:,3,3,2,2).*d2uyzy2 + d1Cz(:,:,:,3,3,2,2).*d1uyy2 + ...
             P.C(:,:,:,3,3,3,3).*d2uzzz2 + d1Cz(:,:,:,3,3,3,3).*d1uzz2 + ...
             P.C(:,:,:,3,1,1,2).*(d2uyxx2 + d2uxxy2) + d1Cx(:,:,:,3,1,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,3,1,1,3).*(d2uzxx2 + d2uxxz2) + d1Cx(:,:,:,3,1,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,3,1,2,3).*(d2uzxy2 + d2uyxz2) + d1Cx(:,:,:,3,1,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,3,2,1,2).*(d2uyyx2 + d2uxyy2) + d1Cy(:,:,:,3,2,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,3,2,1,3).*(d2uzyx2 + d2uxyz2) + d1Cy(:,:,:,3,2,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,3,2,2,3).*(d2uzyy2 + d2uyyz2) + d1Cy(:,:,:,3,2,2,3).*(d1uzy2 + d1uyz2) + ...
             P.C(:,:,:,3,3,1,2).*(d2uyzx2 + d2uxzy2) + d1Cz(:,:,:,3,3,1,2).*(d1uyx2 + d1uxy2) + ...
             P.C(:,:,:,3,3,1,3).*(d2uzzx2 + d2uxzz2) + d1Cz(:,:,:,3,3,1,3).*(d1uzx2 + d1uxz2) + ...
             P.C(:,:,:,3,3,2,3).*(d2uzzy2 + d2uyzz2) + d1Cz(:,:,:,3,3,2,3).*(d1uzy2 + d1uyz2);
   case 7;	%	1D, Anelastic
      rhsX = P.N(:,1,3,3,3)'.*d3uzzzt + d1Nz(:,1,3,3,3)'.*d2uzzt + ...
             P.N(:,1,3,1,3)'.*d3uxzzt + d1Nz(:,1,3,1,3)'.*d2uxzt + ...
             P.N(:,1,3,2,3)'.*d3uyzzt + d1Nz(:,1,3,2,3)'.*d2uyzt;
      rhsY = P.N(:,2,3,3,3)'.*d3uzzzt + d1Nz(:,2,3,3,3)'.*d2uzzt + ...
             P.N(:,2,3,1,3)'.*d3uxzzt + d1Nz(:,2,3,1,3)'.*d2uxzt + ...
             P.N(:,2,3,2,3)'.*d3uyzzt + d1Nz(:,2,3,2,3)'.*d2uyzt;
      rhsZ = P.N(:,3,3,3,3)'.*d3uzzzt + d1Nz(:,3,3,3,3)'.*d2uzzt + ...
             P.N(:,3,3,1,3)'.*d3uxzzt + d1Nz(:,3,3,1,3)'.*d2uxzt + ...
             P.N(:,3,3,2,3)'.*d3uyzzt + d1Nz(:,3,3,2,3)'.*d2uyzt;
   case 8;	%	2D, Anelastic
      rhsX = P.N(:,:,1,1,1,1).*d3uxxxt + d1Nx(:,:,1,1,1,1).*d2uxxt + ...
             P.N(:,:,1,1,3,3).*d3uzxzt + d1Nx(:,:,1,1,3,3).*d2uzzt + ...
             P.N(:,:,1,3,1,1).*d3uxzxt + d1Nz(:,:,1,3,1,1).*d2uxxt + ...
             P.N(:,:,1,3,3,3).*d3uzzzt + d1Nz(:,:,1,3,3,3).*d2uzzt + ...
             P.N(:,:,1,1,1,2).*(d3uyxxt) +            d1Nx(:,:,1,1,1,2).*(d2uyxt) + ...
             P.N(:,:,1,1,1,3).*(d3uzxxt  + d3uxxzt) + d1Nx(:,:,1,1,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,1,1,2,3).*(d3uyxzt) +            d1Nx(:,:,1,1,2,3).*(d2uyzt) + ...
             P.N(:,:,1,3,1,2).*(d3uyzxt) +            d1Nz(:,:,1,3,1,2).*(d2uyxt) + ...
             P.N(:,:,1,3,1,3).*(d3uzzxt  + d3uxzzt) + d1Nz(:,:,1,3,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,1,3,2,3).*(d3uyzzt) +            d1Nz(:,:,1,3,2,3).*(d2uyzt);
      rhsY = P.N(:,:,2,1,1,1).*d3uxxxt + d1Nx(:,:,2,1,1,1).*d2uxxt + ...
             P.N(:,:,2,1,3,3).*d3uzxzt + d1Nx(:,:,2,1,3,3).*d2uzzt + ...
             P.N(:,:,2,3,1,1).*d3uxzxt + d1Nz(:,:,2,3,1,1).*d2uxxt + ...
             P.N(:,:,2,3,3,3).*d3uzzzt + d1Nz(:,:,2,3,3,3).*d2uzzt + ...
             P.N(:,:,2,1,1,2).*(d3uyxxt) +            d1Nx(:,:,2,1,1,2).*(d2uyxt) + ...
             P.N(:,:,2,1,1,3).*(d3uzxxt  + d3uxxzt) + d1Nx(:,:,2,1,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,2,1,2,3).*(d3uyxzt) +            d1Nx(:,:,2,1,2,3).*(d2uyzt) + ...
             P.N(:,:,2,3,1,2).*(d3uyzxt) +            d1Nz(:,:,2,3,1,2).*(d2uyxt) + ...
             P.N(:,:,2,3,1,3).*(d3uzzxt  + d3uxzzt) + d1Nz(:,:,2,3,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,2,3,2,3).*(d3uyzzt) +            d1Nz(:,:,2,3,2,3).*(d2uyzt);
      rhsZ = P.N(:,:,3,1,1,1).*d3uxxxt + d1Nx(:,:,3,1,1,1).*d2uxxt + ...
             P.N(:,:,3,1,3,3).*d3uzxzt + d1Nx(:,:,3,1,3,3).*d2uzzt + ...
             P.N(:,:,3,3,1,1).*d3uxzxt + d1Nz(:,:,3,3,1,1).*d2uxxt + ...
             P.N(:,:,3,3,3,3).*d3uzzzt + d1Nz(:,:,3,3,3,3).*d2uzzt + ...
             P.N(:,:,3,1,1,2).*(d3uyxxt) +            d1Nx(:,:,3,1,1,2).*(d2uyxt) + ...
             P.N(:,:,3,1,1,3).*(d3uzxxt  + d3uxxzt) + d1Nx(:,:,3,1,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,3,1,2,3).*(d3uyxzt) +            d1Nx(:,:,3,1,2,3).*(d2uyzt) + ...
             P.N(:,:,3,3,1,2).*(d3uyzxt) +            d1Nz(:,:,3,3,1,2).*(d2uyxt) + ...
             P.N(:,:,3,3,1,3).*(d3uzzxt  + d3uxzzt) + d1Nz(:,:,3,3,1,3).*(d2uzxt  + d2uxzt) + ...
             P.N(:,:,3,3,2,3).*(d3uyzzt) +            d1Nz(:,:,3,3,2,3).*(d2uyzt);
   case 9;	%	3D, Anelastic
      rhsX = P.N(:,:,:,1,1,1,1).*d3uxxxt + d1Nx(:,:,:,1,1,1,1).*d2uxxt + ...
             P.N(:,:,:,1,1,2,2).*d3uyxyt + d1Nx(:,:,:,1,1,2,2).*d2uyyt + ...
             P.N(:,:,:,1,1,3,3).*d3uzxzt + d1Nx(:,:,:,1,1,3,3).*d2uzzt + ...
             P.N(:,:,:,1,2,1,1).*d3uxyxt + d1Ny(:,:,:,1,2,1,1).*d2uxxt + ...
             P.N(:,:,:,1,2,2,2).*d3uyyyt + d1Ny(:,:,:,1,2,2,2).*d2uyyt + ...
             P.N(:,:,:,1,2,3,3).*d3uzyzt + d1Ny(:,:,:,1,2,3,3).*d2uzzt + ...
             P.N(:,:,:,1,3,1,1).*d3uxzxt + d1Nz(:,:,:,1,3,1,1).*d2uxxt + ...
             P.N(:,:,:,1,3,2,2).*d3uyzyt + d1Nz(:,:,:,1,3,2,2).*d2uyyt + ...
             P.N(:,:,:,1,3,3,3).*d3uzzzt + d1Nz(:,:,:,1,3,3,3).*d2uzzt + ...
             P.N(:,:,:,1,1,1,2).*(d3uyxxt + d3uxxyt) + d1Nx(:,:,:,1,1,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,1,1,1,3).*(d3uzxxt + d3uxxzt) + d1Nx(:,:,:,1,1,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,1,1,2,3).*(d3uzxyt + d3uyxzt) + d1Nx(:,:,:,1,1,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,1,2,1,2).*(d3uyyxt + d3uxyyt) + d1Ny(:,:,:,1,2,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,1,2,1,3).*(d3uzyxt + d3uxyzt) + d1Ny(:,:,:,1,2,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,1,2,2,3).*(d3uzyyt + d3uyyzt) + d1Ny(:,:,:,1,2,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,1,3,1,2).*(d3uyzxt + d3uxzyt) + d1Nz(:,:,:,1,3,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,1,3,1,3).*(d3uzzxt + d3uxzzt) + d1Nz(:,:,:,1,3,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,1,3,2,3).*(d3uzzyt + d3uyzzt) + d1Nz(:,:,:,1,3,2,3).*(d2uzyt + d2uyzt);
      rhsY = P.N(:,:,:,2,1,1,1).*d3uxxxt + d1Nx(:,:,:,2,1,1,1).*d2uxxt + ...
             P.N(:,:,:,2,1,2,2).*d3uyxyt + d1Nx(:,:,:,2,1,2,2).*d2uyyt + ...
             P.N(:,:,:,2,1,3,3).*d3uzxzt + d1Nx(:,:,:,2,1,3,3).*d2uzzt + ...
             P.N(:,:,:,2,2,1,1).*d3uxyxt + d1Ny(:,:,:,2,2,1,1).*d2uxxt + ...
             P.N(:,:,:,2,2,2,2).*d3uyyyt + d1Ny(:,:,:,2,2,2,2).*d2uyyt + ...
             P.N(:,:,:,2,2,3,3).*d3uzyzt + d1Ny(:,:,:,2,2,3,3).*d2uzzt + ...
             P.N(:,:,:,2,3,1,1).*d3uxzxt + d1Nz(:,:,:,2,3,1,1).*d2uxxt + ...
             P.N(:,:,:,2,3,2,2).*d3uyzyt + d1Nz(:,:,:,2,3,2,2).*d2uyyt + ...
             P.N(:,:,:,2,3,3,3).*d3uzzzt + d1Nz(:,:,:,2,3,3,3).*d2uzzt + ...
             P.N(:,:,:,2,1,1,2).*(d3uyxxt + d3uxxyt) + d1Nx(:,:,:,2,1,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,2,1,1,3).*(d3uzxxt + d3uxxzt) + d1Nx(:,:,:,2,1,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,2,1,2,3).*(d3uzxyt + d3uyxzt) + d1Nx(:,:,:,2,1,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,2,2,1,2).*(d3uyyxt + d3uxyyt) + d1Ny(:,:,:,2,2,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,2,2,1,3).*(d3uzyxt + d3uxyzt) + d1Ny(:,:,:,2,2,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,2,2,2,3).*(d3uzyyt + d3uyyzt) + d1Ny(:,:,:,2,2,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,2,3,1,2).*(d3uyzxt + d3uxzyt) + d1Nz(:,:,:,2,3,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,2,3,1,3).*(d3uzzxt + d3uxzzt) + d1Nz(:,:,:,2,3,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,2,3,2,3).*(d3uzzyt + d3uyzzt) + d1Nz(:,:,:,2,3,2,3).*(d2uzyt + d2uyzt);
      rhsZ = P.N(:,:,:,3,1,1,1).*d3uxxxt + d1Nx(:,:,:,3,1,1,1).*d2uxxt + ...
             P.N(:,:,:,3,1,2,2).*d3uyxyt + d1Nx(:,:,:,3,1,2,2).*d2uyyt + ...
             P.N(:,:,:,3,1,3,3).*d3uzxzt + d1Nx(:,:,:,3,1,3,3).*d2uzzt + ...
             P.N(:,:,:,3,2,1,1).*d3uxyxt + d1Ny(:,:,:,3,2,1,1).*d2uxxt + ...
             P.N(:,:,:,3,2,2,2).*d3uyyyt + d1Ny(:,:,:,3,2,2,2).*d2uyyt + ...
             P.N(:,:,:,3,2,3,3).*d3uzyzt + d1Ny(:,:,:,3,2,3,3).*d2uzzt + ...
             P.N(:,:,:,3,3,1,1).*d3uxzxt + d1Nz(:,:,:,3,3,1,1).*d2uxxt + ...
             P.N(:,:,:,3,3,2,2).*d3uyzyt + d1Nz(:,:,:,3,3,2,2).*d2uyyt + ...
             P.N(:,:,:,3,3,3,3).*d3uzzzt + d1Nz(:,:,:,3,3,3,3).*d2uzzt + ...
             P.N(:,:,:,3,1,1,2).*(d3uyxxt + d3uxxyt) + d1Nx(:,:,:,3,1,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,3,1,1,3).*(d3uzxxt + d3uxxzt) + d1Nx(:,:,:,3,1,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,3,1,2,3).*(d3uzxyt + d3uyxzt) + d1Nx(:,:,:,3,1,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,3,2,1,2).*(d3uyyxt + d3uxyyt) + d1Ny(:,:,:,3,2,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,3,2,1,3).*(d3uzyxt + d3uxyzt) + d1Ny(:,:,:,3,2,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,3,2,2,3).*(d3uzyyt + d3uyyzt) + d1Ny(:,:,:,3,2,2,3).*(d2uzyt + d2uyzt) + ...
             P.N(:,:,:,3,3,1,2).*(d3uyzxt + d3uxzyt) + d1Nz(:,:,:,3,3,1,2).*(d2uyxt + d2uxyt) + ...
             P.N(:,:,:,3,3,1,3).*(d3uzzxt + d3uxzzt) + d1Nz(:,:,:,3,3,1,3).*(d2uzxt + d2uxzt) + ...
             P.N(:,:,:,3,3,2,3).*(d3uzzyt + d3uyzzt) + d1Nz(:,:,:,3,3,2,3).*(d2uzyt + d2uyzt);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%		GetRetVar					%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function GetRetVar(nd,t,skip,P)

global u f fr ...
       ux2 ux3 ...
       uy2 uy3 ...
       uz2 uz3;

t = double(t);
skip = double(skip);
nt = double(P.nt);

if P.nx == 1
   if nd > 1
      for i = 1:nd
         u.ux(i,t) = ux2(P.ud(i));
         u.uy(i,t) = uy2(P.ud(i));
         u.uz(i,t) = uz2(P.ud(i));
      end
   else
      u.ux(t) = ux2(P.ud);
      u.uy(t) = uy2(P.ud);
      u.uz(t) = uz2(P.ud);
   end
   
   if mod(t/skip,1) == 0
      fprintf('Time: %7.3fs. Percent complete: %5.2f\r',t*P.dt,100*t/nt);
      if isempty(f) == 0
         fr = fr + 1;
         f(:,:,fr) = [P.X*ux2(:) ...
                      P.Y*uy2(:)+P.Y ...
                      P.Z*uz2(:)+P.Y+P.Z]';
         %plot(f(:,:,fr)')
         clf; hold on
         plot(squeeze(f(1,:,fr)),P.dz*(1:P.nz),'r')
         plot(squeeze(f(2,:,fr)),P.dz*(1:P.nz),'b')
         plot(squeeze(f(3,:,fr)),P.dz*(1:P.nz),'g')
         legend('X-Disp','Y-Disp','Z-Disp')
         box on
         set(gca,'ydir','reverse')
         xlim([-2 4])
         pause(.01)
      end
   end
elseif P.ny == 1
   if nd > 1
      for i = 1:nd
         u.ux(i,:,t) = ux2(:,P.ud(i));
         u.uy(i,:,t) = uy2(:,P.ud(i));
         u.uz(i,:,t) = uz2(:,P.ud(i));
      end
   else
      u.ux(:,t) = ux2(:,P.ud);
      u.uy(:,t) = uy2(:,P.ud);
      u.uz(:,t) = uz2(:,P.ud);
   end
   
   if mod(t/skip,1) == 0
      fprintf('Time: %7.3fs. Percent complete: %5.2f\r',t*P.dt,100*t/nt);
      if isempty(f) == 0
         X = P.dx*P.nx;
         Z = P.dz*P.nz;
         xvec = 0:P.dx:(3*X-P.dx);
         zvec = 0:P.dz:(Z-P.dz);
         fr = fr + 1;
         f(:,:,fr) = [P.X*ux2(:,:)' ...
                      P.Y*uy2(:,:)' ...
                      P.Z*uz2(:,:)']';
         pcolor(xvec,zvec,f(:,:,fr)')
         line([X+P.dx X+P.dx],[0 Z],'color','k')
         line([2*X+P.dx 2*X+P.dx],[0 Z],'color','k')
         text(X/2,0,'x-displacement','horizontalalignment','right','rotation',-90)
         text(X+X/2,0,'y-disp','horizontalalignment','right','rotation',-90)
         text(2*X+X/2,0,'z-disp','horizontalalignment','right','rotation',-90)
         set(gca,'ydir','reverse')
         set(gca,'xtick',[X 2*X])
         daspect([1 1 1])
         shading flat
         caxis([-1 1])
         colorbar
         pause(.01)
      end
   end
else
   if nd > 1
      for i = 1:nd
         u.ux(i,:,t) = ux2(:,round(P.ny/2),P.ud(i));
         u.uy(i,:,t) = uy2(:,round(P.ny/2),P.ud(i));
         u.uz(i,:,t) = uz2(:,round(P.ny/2),P.ud(i));
      end
   else
      u.ux(:,:,t) = ux2(:,:,P.ud);
      u.uy(:,:,t) = uy2(:,:,P.ud);
      u.uz(:,:,t) = uz2(:,:,P.ud);
   end
   
   if mod(t/skip,1) == 0
      fprintf('Time: %7.3fs. Percent complete: %5.2f\r',t*P.dt,100*t/nt);
      if isempty(f) == 0
         X = P.dx*P.nx;
         Z = P.dz*P.nz;
         xvec = 0:P.dx:(3*X-P.dx);
         zvec = 0:P.dz:(Z-P.dz);
         fr = fr + 1;
         f(:,:,fr) = [squeeze(P.X*ux2(:,round(P.ny/2),:))' ...
               squeeze(P.Y*uy2(:,round(P.ny/2),:))' ...
               squeeze(P.Z*uz2(:,round(P.ny/2),:))']';
         pcolor(xvec,zvec,f(:,:,fr)')
         line([X+P.dx X+P.dx],[0 Z],'color','k')
         line([2*X+P.dx 2*X+P.dx],[0 Z],'color','k')
         text(X/2,0,'x-disp','horizontalalignment','right','rotation',-90)
         text(X+X/2,0,'y-disp','horizontalalignment','right','rotation',-90)
         text(2*X+X/2,0,'z-disp','horizontalalignment','right','rotation',-90)
         set(gca,'ydir','reverse')
         set(gca,'xtick',[X 2*X])
         daspect([1 1 1])
         shading flat
         caxis([-1 1])
         colorbar
         pause(.01)
      end
   end
end
