%	Par.m used to populate P structure for use in FDWaveAni.m It requires the
%	subroutines Shift, CProp.m, and Gauss.m
%
%	Model Specific Parameters
%	wType - incoming wave type
%		1 - P-wave
%		2 - radial S-wave
%		3 - transverse S-wave
%	dx, dy, and dz are the node spacings in the x,y, and z directions.
%	nx, ny, and nz are the number of nodes in the x, y, and z directions.
%		if ny = 1, then the problem is 2D.
%	nt is the number of time points to calculate.
%	dt is the sampling interval.
%	T is the dominant period of the source waveform.
%	S is the slowness.
%	az is the backazimuth
%
clear P

P.wType = 3; % Set wave type to P-wave (1), radial shear wave (2), or transverse shear wave (3)

%P.dx = 500;
%P.dy = 1000;
%P.dz = 500;
%P.nx = 10;
%P.ny = 1;
%P.nz = 100;
P.dx = 250;
P.dy = 1000;
P.dz = 250;
P.nx = 5;
P.ny = 1;
P.nz = 200;
P.dt = 0.01;
P.T = 1;
P.S = 0;
P.az = 270;
P.Q = 1e10;
P.Iso = 1;
P.Ela = 0;

%
%	Material parameters
%

%PREM
EP = EarthParam([P.dz:P.dz:P.dz*P.nz]/1e3);
%This works for 1, 2 and 3D
d = ones(P.ny,1)*EP.Density;
for j = 1:P.nx
	P.D(j,:,:) = d;
end
P.C = CProp(P,0,0,1:P.nx,1:P.ny,1:P.nz);	% Isotropic PREM
%P.D = 3300*ones(P.nx,P.ny,P.nz);
%P.C = CProp(P,1,0,1:P.nx,1:P.ny,1:P.nz);	% Isotropic Olivine
%P.C = CProp(P,1,1,1:P.nx,1:P.ny,1:P.nz);	% Anisotropic Olivine
if P.Ela == 0
	P.N = (P.C*2*P.T/(2*pi))*(1/P.Q);
end

%
%	Boundary conditions
%
%	Set up boundary conditions
%	0 - fixed boundary
%	1 - free surface
%	2 - absorbing boundary
%	3 - periodic boundary, only applied to vertical faces.
%

P.BCxy1 = 1;	% BCxy1 applies in 1D, 2D, and 3D problems
P.BCxy2 = 2;	% BCxy2 applies in 1D, 2D, and 3D problems
P.BCyz1 = 3;	% BCyz1 applies in 2D, and 3D problems
P.BCyz2 = 3;	% BCyz2 applies in 2D, and 3D problems
P.BCxz1 = 3;	% BCxz1 applies in 3D problems
P.BCxz2 = 3;	% BCxz2 applies in 3D problems

if P.wType == 1
%	For P-waves
	BotVel = sqrt(mean([P.C(P.nx,P.ny,P.nz,1,1,1,1) P.C(P.nx,P.ny,P.nz,2,2,2,2) P.C(P.nx,P.ny,P.nz,3,3,3,3)])/P.D(P.nx,P.ny,P.nz));
	SideVel = sqrt(mean([P.C(P.nx,P.ny,P.nz,1,1,1,1) P.C(P.nx,P.ny,P.nz,2,2,2,2) P.C(P.nx,P.ny,P.nz,3,3,3,3)])/P.D(P.nx,P.ny,P.nz));
else
%	For S-waves
	BotVel = sqrt(mean([P.C(P.nx,P.ny,P.nz,2,3,2,3) P.C(P.nx,P.ny,P.nz,1,3,1,3) P.C(P.nx,P.ny,P.nz,1,2,1,2)])/P.D(P.nx,P.ny,P.nz));
	SideVel = sqrt(mean([P.C(P.nx,P.ny,P.nz,2,3,2,3) P.C(P.nx,P.ny,P.nz,1,3,1,3) P.C(P.nx,P.ny,P.nz,1,2,1,2)])/P.D(P.nx,P.ny,P.nz));
end

% Azimuth
az = P.az * pi/180;
% Incidence angle of ray with respect to vertical
an = asin(BotVel*P.S);
Sh = sin(an)/BotVel;
Sv = cos(an)/BotVel;
Sx = Sh * sin(az);
Sy = Sh * cos(az);

P.AbVel = 1/abs(Sv);
P.AbVel2 = 1/abs(Sh);
if P.AbVel2 > 1e9
	P.AbVel2 = 1e9;
end

%	Set up initial wavefield

P.ITx = 1:P.nx; nITx = length(P.ITx);
P.ITy = 1:P.ny; nITy = length(P.ITy);
P.ITz = P.nz;   nITz = 1;
len = sqrt((P.nx*P.dx)^2+(P.ny*P.dy)^2);
P.ITnt = round((2*P.T+len*abs(Sh))/P.dt) + 1; %T - total time of initial series
P.ITAx = zeros(nITx,nITy,nITz,P.ITnt);
P.ITAy = zeros(nITx,nITy,nITz,P.ITnt);
P.ITAz = zeros(nITx,nITy,nITz,P.ITnt);
curve = Gauss(P.dt,P.T);
npts = length(curve);
if P.az < 90
	yvec = 1:nITy; xvec = 1:nITx;
elseif P.az < 180
	yvec = nITy:-1:1; xvec = 1:nITx;
elseif P.az < 270
	yvec = nITy:-1:1; xvec = nITx:-1:1;
else
	yvec = 1:nITy; xvec = nITx:-1:1;
end
switch P.wType
    case 1
	% P-wave
	P.ITAx(xvec(1),yvec(1),1,1:npts) = sin(az)*sin(an)*curve;
	P.ITAy(xvec(1),yvec(1),1,1:npts) = cos(az)*sin(an)*curve;
	P.ITAz(xvec(1),yvec(1),1,1:npts) = -cos(an)*curve;
    case 2
	% S-wave, radial
	P.ITAx(xvec(1),yvec(1),1,1:npts) = sin(az)*cos(an)*curve;
	P.ITAy(xvec(1),yvec(1),1,1:npts) = cos(az)*cos(an)*curve;
	P.ITAz(xvec(1),yvec(1),1,1:npts) = -sin(an)*curve;
    case 3
	% S-wave, transverse
	P.ITAx(xvec(1),yvec(1),1,1:npts) = cos(az)*curve;
	P.ITAy(xvec(1),yvec(1),1,1:npts) = sin(az)*curve;
	P.ITAz(xvec(1),yvec(1),1,1:npts) = 0;
end

nx = 0;
for j = xvec
	nx = nx + 1;
	ny = 0;
	for k = yvec
		ny = ny + 1;
		t = (nx-1)*P.dx*abs(Sx) + (ny-1)*P.dy*abs(Sy);
		P.ITAx(j,k,1,:) = Shift(squeeze(P.ITAx(xvec(1),yvec(1),1,:))',P.dt,t);
		P.ITAy(j,k,1,:) = Shift(squeeze(P.ITAy(xvec(1),yvec(1),1,:))',P.dt,t);
		P.ITAz(j,k,1,:) = Shift(squeeze(P.ITAz(xvec(1),yvec(1),1,:))',P.dt,t);
	end
end

%
%	Output Parameters
%
%	ud are the depths (in terms of index) at which time series are returned.
%	fr is the displacement component for which the frame is captured. 
%	nf is the number of frames to capture. If nf is equal to 0, no frames
%		are captured and every 10th timestep is displayed.
%	fig is the figure to which the wavefield is displayed.
%	0 - x
%	1 - y
%	2 - z
%	X, Y, and Z are the amplification factors used in the display of the wavefield.
%

P.ud = [1];
P.fr = 2;
P.nf = 200;
P.fig = 11;
P.X = 1;
P.Y = 1;
P.Z = 1;

%	Redefine some parameters for convenience.

P.nt = int32(P.ITnt + 2*P.dz*P.nz/BotVel/P.dt);

if P.Iso == 1
	P.L = P.C(:,:,:,1,1,3,3);
	P.U = P.C(:,:,:,1,3,1,3);
	P.L = squeeze(P.L);
	P.U = squeeze(P.U);
end
P.D = squeeze(P.D);
P.C = squeeze(P.C);
if P.Ela == 0
	P.N = squeeze(P.N);
end
if P.nx == 1
	P.D = P.D';
	if P.Iso == 1
		P.L = P.L';
		P.U = P.U';
	end
end
